from __future__ import annotations

import re
import sys
from itertools import chain
from typing import Annotated, Any, Literal, NamedTuple

import zhon
from more_itertools import batched
from pydantic import (
    Field,
    FieldSerializationInfo,
    RootModel,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)
from retrie.retrie import Blacklist
from typing_extensions import Self

from libresvip.core.constants import DEFAULT_PHONEME
from libresvip.core.time_interval import RangeInterval
from libresvip.model.base import BaseModel, Note
from libresvip.model.point import PointList
from libresvip.utils.audio import audio_path_validator
from libresvip.utils.text import uuid_str

from . import constants
from .interval_utils import position_to_ticks

symbols_blacklist = Blacklist(
    [
        "(",
        ")",
        "[",
        "]",
        "{",
        "}",
        "（",
        "）",
        "<",
        ">",
        "《",
        "》",
        "―",
        "—",
        "*",
        "×",
        "!",
        "！",
        "?",
        "？",
        ":",
        "：",
        "·",
        "•",
        "。",
        ",",
        "，",
        ";",
        "；",
        "^",
        "`",
        '"',
        "‘",
        "’",
        "“",
        "”",
        "=",
        "、",
        "_",
        "$",
        "%",
        "~",
        "@",
        "#",
        "…",
        "&",
        "￥",
    ],
    match_substrings=True,
)


class SVPoint(NamedTuple):
    offset: int
    value: float


class SVPoints(PointList[SVPoint], RootModel[list[SVPoint]]):
    root: list[SVPoint] = Field(default_factory=list)


class SVBaseAttributes(BaseModel):
    t_f0_left: float | None = Field(None, alias="tF0Left")
    t_f0_right: float | None = Field(None, alias="tF0Right")
    d_f0_left: float | None = Field(None, alias="dF0Left")
    d_f0_right: float | None = Field(None, alias="dF0Right")
    t_f0_vbr_start: float | None = Field(None, alias="tF0VbrStart")
    t_f0_vbr_left: float | None = Field(None, alias="tF0VbrLeft")
    t_f0_vbr_right: float | None = Field(None, alias="tF0VbrRight")
    d_f0_vbr: float | None = Field(None, alias="dF0Vbr")
    f_f0_vbr: float | None = Field(None, alias="fF0Vbr")
    param_loudness: float | None = Field(None, alias="paramLoudness")
    param_tension: float | None = Field(None, alias="paramTension")
    param_breathiness: float | None = Field(None, alias="paramBreathiness")
    param_gender: float | None = Field(None, alias="paramGender")
    param_tone_shift: float | None = Field(None, alias="paramToneShift")
    improvise_attack_release: bool | None = Field(None, alias="improviseAttackRelease")
    language_override: str | None = Field(None, alias="languageOverride")
    phoneset_override: str | None = Field(None, alias="phonesetOverride")


class SVMeter(BaseModel):
    index: int
    denominator: int = 4
    numerator: int = 4


class SVTempo(BaseModel):
    bpm: float
    position: int


class SVTime(BaseModel):
    meter: list[SVMeter] = Field(default_factory=list)
    tempo: list[SVTempo] = Field(default_factory=list)


class SVBasePoints(BaseModel):
    points: SVPoints = Field(default_factory=SVPoints)

    @field_validator("points", mode="before")
    @classmethod
    def validate_points(cls, points: list[float], _info: ValidationInfo) -> SVPoints:
        if _info.mode == "json":
            return SVPoints(root=[SVPoint._make(each) for each in batched(points, 2)])
        return SVPoints(root=points)

    @field_serializer("points", when_used="json")
    def serialize_points(self, points: SVPoints, _info: FieldSerializationInfo) -> list[float]:
        return list(chain.from_iterable(points.root))


class SVParamCurve(SVBasePoints):
    mode: Literal["linear", "cubic", "cosine", "sigmoid"] = Field("linear")

    def edited_range(self, default_value: float = 0.0) -> RangeInterval:
        tolerance = 1e-6
        interval = RangeInterval()
        points = [
            SVPoint(position_to_ticks(point.offset), point.value) for point in self.points.root
        ]
        if not points:
            return interval
        elif len(points) == 1:
            return (
                interval
                if abs(points[0].value - default_value) < tolerance
                else RangeInterval([(0, sys.maxsize // 2)])
            )
        if abs(points[0].value - default_value) > tolerance:
            interval |= RangeInterval([(0, points[0].offset)])
        start, end = points[0].offset, points[0].offset
        for i in range(1, len(points)):
            if (
                abs(points[i - 1].value - default_value) < tolerance
                and abs(points[i].value - default_value) < tolerance
            ):
                if start < end:
                    interval |= RangeInterval([(start, end)])
                start = points[i].offset
            else:
                end = points[i].offset
        if start < end:
            interval |= RangeInterval([(start, end)])
        if abs(points[-1].value - default_value) > tolerance:
            interval |= RangeInterval([(points[-1].offset, sys.maxsize // 2)])
        return interval

    def __add__(self, offset: int) -> SVParamCurve:
        new_curve = self.model_copy(deep=True)
        for i in range(len(new_curve.points)):
            new_curve.points[i] = new_curve.points[i]._replace(
                offset=new_curve.points[i].offset + offset
            )
        return new_curve


class SVParamTake(BaseModel):
    id_value: int = Field(alias="id")
    seed_duration: int | None = Field(None, alias="seedDuration")
    seed_pitch: int | None = Field(None, alias="seedPitch")
    seed_timbre: int | None = Field(None, alias="seedTimbre")
    expr: float | None = None
    liked: bool = False


class SVBasePitchControl(BaseModel):
    pos: int
    pitch: float
    id_value: str = Field(alias="id")


class SVPointPitchControl(SVBasePitchControl):
    type_: Literal["point"] = Field("point", alias="type")


class SVCurvePitchControl(SVBasePitchControl, SVBasePoints):
    type_: Literal["curve"] = Field("curve", alias="type")


SVPitchControl = Annotated[
    SVPointPitchControl | SVCurvePitchControl,
    Field(discriminator="type_"),
]


class SVParamTakes(BaseModel):
    active_take_id: int = Field(alias="activeTakeId")
    takes: list[SVParamTake] = Field(default_factory=list)


class SVPhonemeAttribute(BaseModel):
    left_offset: float | None = Field(None, alias="leftOffset")
    strength: float = 0
    alt: int = 0
    position: float | None = None
    activity: float | None = None


class SVNoteAttributes(SVBaseAttributes):
    t_f0_left: float | None = Field(None, alias="tF0Left")
    t_f0_right: float | None = Field(None, alias="tF0Right")
    d_f0_left: float | None = Field(None, alias="dF0Left")
    d_f0_right: float | None = Field(None, alias="dF0Right")
    t_f0_vbr_start: float | None = Field(None, alias="tF0VbrStart")
    t_f0_vbr_left: float | None = Field(None, alias="tF0VbrLeft")
    t_f0_vbr_right: float | None = Field(None, alias="tF0VbrRight")
    d_f0_vbr: float | None = Field(None, alias="dF0Vbr")
    f_f0_vbr: float | None = Field(None, alias="fF0Vbr")
    t_f0_offset: float | None = Field(None, alias="tF0Offset")
    p_f0_vbr: float | None = Field(None, alias="pF0Vbr")
    d_f0_jitter: float | None = Field(None, alias="dF0Jitter")
    t_note_offset: float | None = Field(None, alias="tNoteOffset")
    dur: list[float] | None = None
    alt: list[int] | None = None
    expr_group: str | None = Field(None, alias="exprGroup")
    exp_value_x: float | None = Field(None, alias="expValueX")
    exp_value_y: float | None = Field(None, alias="expValueY")
    strength: list[float] | None = None
    r_tone: float | None = Field(None, alias="rTone")
    r_intonation: float | None = Field(None, alias="rIntonation")
    even_syllable_duration: float | None = Field(None, alias="evenSyllableDuration")
    phonemes: list[SVPhonemeAttribute] | None = None

    def _get_transition_offset(self) -> float:
        return constants.DEFAULT_PITCH_TRANSITION if self.t_f0_offset is None else self.t_f0_offset

    def _set_transition_offset(self, value: float) -> None:
        self.t_f0_offset = value

    transition_offset = property(_get_transition_offset, _set_transition_offset)

    def _get_portamento_left(self) -> float:
        return constants.DEFAULT_PITCH_PORTAMENTO if self.t_f0_left is None else self.t_f0_left

    def _set_portamento_left(self, value: float) -> None:
        self.t_f0_left = value

    portamento_left = property(_get_portamento_left, _set_portamento_left)

    def _get_portamento_right(self) -> float:
        return constants.DEFAULT_PITCH_PORTAMENTO if self.t_f0_right is None else self.t_f0_right

    def _set_portamento_right(self, value: float) -> None:
        self.t_f0_right = value

    portamento_right = property(_get_portamento_right, _set_portamento_right)

    def _get_depth_left(self) -> float:
        return constants.DEFAULT_PITCH_DEPTH if self.d_f0_left is None else self.d_f0_left

    def _set_depth_left(self, value: float) -> None:
        self.d_f0_left = value

    depth_left = property(_get_depth_left, _set_depth_left)

    def _get_depth_right(self) -> float:
        return constants.DEFAULT_PITCH_DEPTH if self.d_f0_right is None else self.d_f0_right

    def _set_depth_right(self, value: float) -> None:
        self.d_f0_right = value

    depth_right = property(_get_depth_right, _set_depth_right)

    def _get_vibrato_start(self) -> float:
        return (
            constants.DEFAULT_VIBRATO_START if self.t_f0_vbr_start is None else self.t_f0_vbr_start
        )

    def _set_vibrato_start(self, value: float) -> None:
        self.t_f0_vbr_start = value

    vibrato_start = property(_get_vibrato_start, _set_vibrato_start)

    def _get_vibrato_left(self) -> float:
        return constants.DEFAULT_VIBRATO_FADE if self.t_f0_vbr_left is None else self.t_f0_vbr_left

    def _set_vibrato_left(self, value: float) -> None:
        self.t_f0_vbr_left = value

    vibrato_left = property(_get_vibrato_left, _set_vibrato_left)

    def _get_vibrato_right(self) -> float:
        return (
            constants.DEFAULT_VIBRATO_FADE if self.t_f0_vbr_right is None else self.t_f0_vbr_right
        )

    def _set_vibrato_right(self, value: float) -> None:
        self.t_f0_vbr_right = value

    vibrato_right = property(_get_vibrato_right, _set_vibrato_right)

    def _get_vibrato_depth(self) -> float:
        return constants.DEFAULT_VIBRATO_DEPTH if self.d_f0_vbr is None else self.d_f0_vbr

    def _set_vibrato_depth(self, value: float) -> None:
        self.d_f0_vbr = value

    vibrato_depth = property(_get_vibrato_depth, _set_vibrato_depth)

    def _get_vibrato_frequency(self) -> float:
        return constants.DEFAULT_VIBRATO_FREQUENCY if self.f_f0_vbr is None else self.f_f0_vbr

    def _set_vibrato_frequency(self, value: float) -> None:
        self.f_f0_vbr = value

    vibrato_frequency = property(_get_vibrato_frequency, _set_vibrato_frequency)

    def _get_vibrato_phase(self) -> float:
        return constants.DEFAULT_VIBRATO_PHASE if self.p_f0_vbr is None else self.p_f0_vbr

    def _set_vibrato_phase(self, value: float) -> None:
        self.p_f0_vbr = value

    vibrato_phase = property(_get_vibrato_phase, _set_vibrato_phase)

    def _get_vibrato_jitter(self) -> float:
        return constants.DEFAULT_VIBRATO_JITTER if self.d_f0_jitter is None else self.d_f0_jitter

    def _set_vibrato_jitter(self, value: float) -> None:
        self.d_f0_jitter = value

    vibrato_jitter = property(_get_vibrato_jitter, _set_vibrato_jitter)

    def set_phone_duration(self, index: int, duration: float) -> None:
        if self.dur is None:
            self.dur = [1.0] * (index + 1)
        elif len(self.dur) <= index:
            self.dur.extend([1.0] * (index - len(self.dur) + 1))
        self.dur[index] = duration

    def pitch_edited(
        self,
        regard_default_vibrato_as_unedited: bool = True,
        consider_instant_pitch_mode: bool = True,
    ) -> bool:
        transition_edited = any(
            x is not None
            for x in [
                self.t_f0_offset,
                self.t_f0_left,
                self.t_f0_right,
                self.d_f0_left,
                self.d_f0_right,
            ]
        )
        if consider_instant_pitch_mode:
            tolerance = 1e-6
            transition_edited &= any(
                x >= tolerance
                for x in (
                    abs(self.portamento_left - constants.DEFAULT_PITCH_PORTAMENTO),
                    abs(self.portamento_right - constants.DEFAULT_PITCH_PORTAMENTO),
                    abs(self.depth_left - constants.DEFAULT_PITCH_DEPTH),
                    abs(self.depth_right - constants.DEFAULT_PITCH_DEPTH),
                )
            )

        vibrato_edited = self.vibrato_depth != 0.0
        if regard_default_vibrato_as_unedited:
            vibrato_edited &= any(
                x is not None
                for x in (
                    self.d_f0_vbr,
                    self.f_f0_vbr,
                    self.p_f0_vbr,
                    self.t_f0_vbr_left,
                    self.t_f0_vbr_right,
                    self.t_f0_vbr_start,
                )
            )
        return transition_edited or vibrato_edited

    def default_language(self, database: SVDatabase) -> str:
        return self.language_override or database.language_override or database.language


class SVNote(BaseModel):
    onset: int
    duration: int
    lyrics: str = ""
    phonemes: str = ""
    pitch: int
    detune: int | None = None
    accent: str | None = None
    attributes: SVNoteAttributes = Field(default_factory=SVNoteAttributes)
    system_attributes: SVNoteAttributes | None = Field(None, alias="systemAttributes")
    takes: SVParamTakes | None = None
    pitch_takes: SVParamTakes | None = Field(None, alias="pitchTakes")
    timbre_takes: SVParamTakes | None = Field(None, alias="timbreTakes")
    musical_type: Literal["singing", "rap"] | None = Field("singing", alias="musicalType")
    instant_mode: bool | None = Field(None, alias="instantMode")

    def merge_attributes(self, attributes: SVNoteAttributes) -> None:
        ori_dict = self.attributes.model_dump(
            by_alias=True,
            exclude_none=True,
            exclude_unset=True,
            exclude_defaults=True,
        )
        new_dict = attributes.model_dump(
            by_alias=True,
            exclude_none=True,
            exclude_unset=True,
            exclude_defaults=True,
        )
        ori_dict.update({k: v for k, v in new_dict.items() if k not in ori_dict})
        self.attributes = SVNoteAttributes.model_validate(ori_dict)

    @model_validator(mode="after")
    def _merge_system_attribute(self) -> Self:
        if self.system_attributes is not None:
            self.merge_attributes(self.system_attributes)
        return self

    def pitch_edited(
        self,
        regard_default_vibrato_as_unedited: bool = True,
        consider_instant_pitch_mode: bool = True,
    ) -> bool:
        return self.attributes.pitch_edited(
            regard_default_vibrato_as_unedited,
            consider_instant_pitch_mode and self.instant_mode is not False,
        )

    def __add__(self, blick_offset: int) -> SVNote:
        return self.model_copy(deep=True, update={"onset": self.onset + blick_offset})

    def __xor__(self, pitch_offset: int) -> SVNote:
        return self.model_copy(deep=True, update={"pitch": self.pitch + pitch_offset})

    @staticmethod
    def normalize_lyric(lyric: str) -> str:
        return symbols_blacklist.cleanse_text(lyric).strip()

    @classmethod
    def normalize_phoneme(cls, note: Note) -> str:
        if note.pronunciation:
            return note.pronunciation
        elif (hanzi := re.search(rf"[{zhon.hanzi.characters}]+", note.lyric)) is not None:
            return hanzi[0]
        elif valid_chars := cls.normalize_lyric(note.lyric):
            return valid_chars.lstrip(".")
        else:
            return DEFAULT_PHONEME


class SVRenderConfig(BaseModel):
    aspiration_format: str = Field("noAspiration", alias="aspirationFormat")
    bit_depth: int = Field(16, alias="bitDepth")
    destination: str = "./"
    filename: str = "untitled"
    num_channels: int = Field(1, alias="numChannels")
    sample_rate: int = Field(44100, alias="sampleRate")
    export_mix_down: bool = Field(True, alias="exportMixDown")
    export_pitch: bool | None = Field(False, alias="exportPitch")


class SVParameters(BaseModel):
    breathiness: SVParamCurve = Field(default_factory=SVParamCurve, title="气声")
    gender: SVParamCurve = Field(default_factory=SVParamCurve, title="性别")
    loudness: SVParamCurve = Field(default_factory=SVParamCurve, title="响度")
    pitch_delta: SVParamCurve = Field(
        default_factory=SVParamCurve, alias="pitchDelta", title="音高偏差"
    )
    tension: SVParamCurve = Field(default_factory=SVParamCurve, title="张力")
    vibrato_env: SVParamCurve = Field(
        default_factory=SVParamCurve, alias="vibratoEnv", title="颤音包络"
    )
    voicing: SVParamCurve = Field(default_factory=SVParamCurve, title="发声")
    tone_shift: SVParamCurve | None = Field(
        default_factory=SVParamCurve, alias="toneShift", title="音区偏移"
    )
    mouth_opening: SVParamCurve | None = Field(
        default_factory=SVParamCurve, alias="mouthOpening", title="口型"
    )

    def __add__(self, offset: int) -> SVParameters:
        new_params = self.model_copy(deep=True)
        for key in new_params.model_fields:
            if (val := getattr(new_params, key, None)) is not None:
                setattr(new_params, key, val + offset)
        return new_params


class SVVoice(SVBaseAttributes):
    t_f0_left: float | None = Field(0.07, alias="tF0Left")
    t_f0_right: float | None = Field(0.07, alias="tF0Right")
    d_f0_left: float | None = Field(0.15, alias="dF0Left")
    d_f0_right: float | None = Field(0.15, alias="dF0Right")
    t_f0_vbr_start: float | None = Field(0.25, alias="tF0VbrStart")
    t_f0_vbr_left: float | None = Field(0.2, alias="tF0VbrLeft")
    t_f0_vbr_right: float | None = Field(0.2, alias="tF0VbrRight")
    d_f0_vbr: float | None = Field(1.0, alias="dF0Vbr")
    f_f0_vbr: float | None = Field(5.5, alias="fF0Vbr")
    vocal_mode_inherited: bool = Field(True, alias="vocalModeInherited")
    vocal_mode_preset: str = Field("", alias="vocalModePreset")
    vocal_mode_params: dict[str, float | dict[str, float]] | None = Field(
        None, alias="vocalModeParams"
    )
    render_mode: str | None = Field(None, alias="renderMode")

    def to_attributes(self) -> SVNoteAttributes:
        voice_dict = self.model_dump(
            by_alias=True,
            exclude_none=True,
            exclude_unset=True,
            exclude={
                "vocal_mode_inherited",
                "vocal_mode_preset",
                "vocal_mode_params",
                "render_mode",
                "improvise_attack_release",
            },
        )
        return SVNoteAttributes.model_validate(voice_dict)


class SVAudio(BaseModel):
    filename: str = ""
    duration: float

    validate_filename = field_validator("filename", mode="before")(audio_path_validator)


class SVDatabase(BaseModel):
    name: str = ""
    language: str = "mandarin"
    phoneset: str = "xsampa"
    version: str | int | None = None
    language_override: str | None = Field(None, alias="languageOverride")
    phoneset_override: str | None = Field(None, alias="phonesetOverride")
    backend_type: str = Field("", alias="backendType")


class SVRef(BaseModel):
    audio: SVAudio | None = None
    blick_absolute_begin: int | None = Field(0, alias="blickAbsoluteBegin")
    blick_absolute_end: int | None = Field(-1, alias="blickAbsoluteEnd")
    blick_offset: int = Field(default=0, alias="blickOffset")
    pitch_offset: int = Field(default=0, alias="pitchOffset")
    takes: SVParamTakes | None = None
    pitch_takes: SVParamTakes | None = Field(None, alias="pitchTakes")
    timbre_takes: SVParamTakes | None = Field(None, alias="timbreTakes")
    database: SVDatabase = Field(default_factory=SVDatabase)
    dictionary: str = ""
    voice: SVVoice = Field(default_factory=SVVoice)
    group_id: str = Field(default_factory=uuid_str, alias="groupID")
    is_instrumental: bool = Field(default=False, alias="isInstrumental")
    system_pitch_delta: SVParamCurve = Field(default_factory=SVParamCurve, alias="systemPitchDelta")


class SVGroup(BaseModel):
    name: str = "main"
    notes: list[SVNote] = Field(default_factory=list)
    parameters: SVParameters = Field(default_factory=SVParameters)
    uuid: str = Field(default_factory=uuid_str)
    vocal_modes: dict[str, SVParamCurve] = Field(default_factory=dict, alias="vocalModes")
    pitch_controls: list[SVPitchControl] = Field(default_factory=list, alias="pitchControls")

    @field_validator("notes", mode="before")
    @classmethod
    def validate_notes(cls, v: list[dict[str, Any]], _info: ValidationInfo) -> list[dict[str, Any]]:
        if _info.mode == "json":
            return [note for note in v if note["onset"] >= 0]
        return v

    def overlapped_with(self, other: SVGroup) -> bool:
        for note in self.notes:
            for other_note in other.notes:
                x = note.onset - (other_note.onset + other_note.duration)
                y = (note.onset + note.duration) - other_note.onset
                if x * y == 0:
                    continue
                tmp = (x < 0) ^ (y > 0)
                if tmp:
                    return True
        return False

    def __add__(self, blick_offset: int) -> SVGroup:
        return self.model_copy(
            deep=True,
            update={
                "notes": [
                    note + blick_offset for note in self.notes if note.onset + blick_offset >= 0
                ],
                "parameters": self.parameters + blick_offset,
            },
        )

    def __xor__(self, pitch_offset: int) -> SVGroup:
        return self.model_copy(
            deep=True, update={"notes": [note ^ pitch_offset for note in self.notes]}
        )


class SVMixer(BaseModel):
    pan: float = 0.0
    mute: bool = False
    solo: bool = False
    display: bool = True
    gain_decibel: float = Field(0.0, alias="gainDecibel")


class SVTrack(BaseModel):
    disp_color: str = Field(default="ff7db235", alias="dispColor")
    disp_order: int = Field(default=0, alias="dispOrder")
    groups: list[SVRef] = Field(default_factory=list)
    main_group: SVGroup = Field(default_factory=SVGroup, alias="mainGroup")
    main_ref: SVRef = Field(default_factory=SVRef, alias="mainRef")
    mixer: SVMixer = Field(default_factory=SVMixer)
    name: str = "Track 1"
    render_enabled: bool = Field(default=True, alias="renderEnabled")


class SVProject(BaseModel):
    library: list[SVGroup] = Field(default_factory=list)
    render_config: SVRenderConfig = Field(default_factory=SVRenderConfig, alias="renderConfig")
    instant_mode_enabled: bool | None = Field(None, alias="instantModeEnabled")
    time_sig: SVTime = Field(default_factory=SVTime, alias="time")
    tracks: list[SVTrack] = Field(default_factory=list)
    version: int = 100



import sys

from loguru import _defaults, _logger
from PySide6.QtCore import (
    QMessageLogContext,
    QtMsgType,
    qFormatLogMessage,
    qInstallMessageHandler,
    qSetMessagePattern,
)

qt_logger = _logger.Logger(
    core=_logger.Core(),
    exception=None,
    depth=0,
    record=False,
    lazy=False,
    colors=False,
    raw=False,
    capture=True,
    patchers=[],
    extra={},
)
if _defaults.LOGURU_AUTOINIT and sys.stderr:
    qt_logger.add(
        sys.stderr,
        format=(
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level: <8}</level> | "
            "<cyan>{extra[prefix]}</cyan> - <level>{message}</level>"
        ),
    )
log_methods = {
    "info": qt_logger.info,
    "warning": qt_logger.warning,
    "critical": qt_logger.critical,
    "fatal": qt_logger.error,
    "debug": qt_logger.debug,
}


def qt_log_handler(message_type: QtMsgType, ctx: QMessageLogContext, message: str) -> None:
    msg_type, file_name, function, line, msg = qFormatLogMessage(message_type, ctx, message).split(
        "\n", maxsplit=5
    )
    message_prefix = (
        f"{file_name}:{line}" if function == "unknown" else f"{file_name}:{function}:{line}"
    )
    with qt_logger.contextualize(prefix=message_prefix):
        msg = msg.removeprefix(message_prefix).removeprefix(":9").removeprefix(": ")
        log_method = log_methods.get(msg_type, qt_logger.info)
        log_method(msg)


def enable_log_handler() -> None:
    qSetMessagePattern("%{type}\n%{file}\n%{function}\n%{line}\n%{message}")
    qInstallMessageHandler(qt_log_handler)



from itertools import chain
from typing import NamedTuple

from more_itertools import batched
from pydantic import (
    Field,
    FieldSerializationInfo,
    RootModel,
    ValidationInfo,
    field_serializer,
    field_validator,
)

from libresvip.core.constants import DEFAULT_BPM, DEFAULT_PHONEME
from libresvip.model.base import BaseModel
from libresvip.model.point import PointList
from libresvip.utils.audio import audio_path_validator


class S5pPoint(NamedTuple):
    offset: float
    value: float


class S5pPoints(PointList[S5pPoint], RootModel[list[S5pPoint]]):
    root: list[S5pPoint] = Field(default_factory=list)


class S5pMeterItem(BaseModel):
    measure: int = 0
    beat_per_measure: int = Field(4, alias="beatPerMeasure")
    beat_granularity: int = Field(4, alias="beatGranularity")


class S5pTempoItem(BaseModel):
    position: int
    beat_per_minute: float = Field(DEFAULT_BPM, alias="beatPerMinute")


class S5pDbDefaults(BaseModel):
    lyric: str | None = DEFAULT_PHONEME
    breathiness: float | None = 0.0
    gender: float | None = 0.0
    tension: float | None = 0.0
    d_f0_vbr: float = Field(0.025, alias="dF0Vbr")
    p_f0_vbr: float = Field(0.0, alias="pF0Vbr")
    t_f0_vbr_left: float = Field(0.15, alias="tF0VbrLeft")
    t_f0_vbr_right: float = Field(0.15, alias="tF0VbrRight")
    t_f0_vbr_start: float = Field(0.25, alias="tF0VbrStart")
    f_f0_vbr: float = Field(5.5, alias="fF0Vbr")
    t_f0_left: float = Field(0.07, alias="tF0Left")
    t_f0_right: float = Field(0.07, alias="tF0Right")
    d_f0_left: float = Field(0.0, alias="dF0Left")
    d_f0_right: float = Field(0.0, alias="dF0Right")
    d_f0_jitter: float = Field(1.0, alias="dF0Jitter")


class S5pNote(BaseModel):
    lyric: str
    onset: int
    duration: int
    comment: str = ""
    pitch: int
    d_f0_vbr: float | None = Field(None, alias="dF0Vbr")
    p_f0_vbr: float | None = Field(None, alias="pF0Vbr")
    t_f0_vbr_left: float | None = Field(None, alias="tF0VbrLeft")
    t_f0_vbr_right: float | None = Field(None, alias="tF0VbrRight")
    t_f0_vbr_start: float | None = Field(None, alias="tF0VbrStart")
    f_f0_vbr: float | None = Field(None, alias="fF0Vbr")
    t_f0_left: float | None = Field(None, alias="tF0Left")
    t_f0_right: float | None = Field(None, alias="tF0Right")
    d_f0_left: float | None = Field(None, alias="dF0Left")
    d_f0_right: float | None = Field(None, alias="dF0Right")
    d_f0_jitter: float | None = Field(None, alias="dF0Jitter")
    t_f0_offset: float | None = Field(None, alias="tF0Offset")
    t_note_offset: float | None = Field(None, alias="tNoteOffset")
    t_syl_onset: float | None = Field(None, alias="tSylOnset")
    t_syl_coda: float | None = Field(None, alias="tSylCoda")
    w_syl_nucleus: float | None = Field(None, alias="wSylNucleus")
    sublib: str | None = None


class S5pTrackMixer(BaseModel):
    gain_decibel: float = Field(0.0, alias="gainDecibel")
    pan: float = 0.0
    muted: bool = False
    solo: bool = False
    engine_on: bool = Field(True, alias="engineOn")
    display: bool = Field(True)


class S5pParameters(BaseModel):
    interval: int = 5512500
    pitch_delta: S5pPoints = Field(default_factory=S5pPoints, alias="pitchDelta")
    vibrato_env: S5pPoints = Field(default_factory=S5pPoints, alias="vibratoEnv")
    loudness: S5pPoints = Field(default_factory=S5pPoints)
    tension: S5pPoints = Field(default_factory=S5pPoints)
    breathiness: S5pPoints = Field(default_factory=S5pPoints)
    voicing: S5pPoints = Field(default_factory=S5pPoints)
    gender: S5pPoints = Field(default_factory=S5pPoints)

    @field_validator(
        "pitch_delta",
        "vibrato_env",
        "loudness",
        "tension",
        "breathiness",
        "voicing",
        "gender",
        mode="before",
    )
    @classmethod
    def validate_points(cls, points: list[float], _info: ValidationInfo) -> S5pPoints:
        if _info.mode == "json":
            return S5pPoints(root=[S5pPoint._make(each) for each in batched(points, 2)])
        return points if isinstance(points, S5pPoints) else S5pPoints(root=points)

    @field_serializer(
        "pitch_delta",
        "vibrato_env",
        "loudness",
        "tension",
        "breathiness",
        "voicing",
        "gender",
        when_used="json",
    )
    def serialize_points(self, points: S5pPoints, _info: FieldSerializationInfo) -> list[float]:
        return list(chain.from_iterable(points.root))


class S5pTrack(BaseModel):
    name: str
    db_name: str = Field("", alias="dbName")
    color: str = "15e879"
    display_order: int = Field(0, alias="displayOrder")
    db_defaults: S5pDbDefaults = Field(default_factory=S5pDbDefaults, alias="dbDefaults")
    notes: list[S5pNote | None] = Field(default_factory=list)
    gs_events: None = Field(None, alias="gsEvents")
    mixer: S5pTrackMixer = Field(default_factory=S5pTrackMixer)
    parameters: S5pParameters = Field(default_factory=S5pParameters)


class S5pInstrumental(BaseModel):
    filename: str = ""
    offset: float = 0.0

    validate_filename = field_validator("filename", mode="before")(audio_path_validator)


class S5pMixer(BaseModel):
    gain_instrumental_decibel: float = Field(0.0, alias="gainInstrumentalDecibel")
    gain_vocal_master_decibel: float = Field(0.0, alias="gainVocalMasterDecibel")
    instrumental_muted: bool = Field(False, alias="instrumentalMuted")
    vocal_master_muted: bool = Field(False, alias="vocalMasterMuted")


class S5pProject(BaseModel):
    version: int = 7
    meter: list[S5pMeterItem] = Field(default_factory=list)
    tempo: list[S5pTempoItem] = Field(default_factory=list)
    tracks: list[S5pTrack] = Field(default_factory=list)
    instrumental: S5pInstrumental = Field(default_factory=S5pInstrumental)
    mixer: S5pMixer = Field(default_factory=S5pMixer)



# Ported from QNrbf by SineStriker
import dataclasses
import pathlib
from typing import Any

from construct import Container
from loguru import logger

from libresvip.utils.translation import gettext_lazy as _

from .binary_models import (
    PrimitiveTypeEnum,
    RecordTypeEnum,
    SerializedStreamHeader,
    SVIPFile,
    references_by_id,
)
from .nrbf_iobase import NrbfIOBase
from .xstudio_models import XSAppModel, fullname2classes


@dataclasses.dataclass
class SvipReader(NrbfIOBase):
    svip_file: SVIPFile = dataclasses.field(init=False)
    xstudio_model: XSAppModel = dataclasses.field(init=False)
    header: SerializedStreamHeader = dataclasses.field(init=False)

    def build_binary_array(self, obj: Container) -> list[Any | None]:
        results: list[Any | None] = []
        if "Class" in str(obj.binary_type_enum):
            if obj.member_values is not None:
                for member in obj.member_values:
                    if member is None or member.real_obj is None or member.real_obj.obj is None:
                        results.append(None)
                    else:
                        obj = member.real_obj.obj
                        if "real_obj" in obj:
                            results.append(self.build_object(obj["real_obj"]))
                        else:
                            results.append(self.build_object(obj))
            else:
                results = [None] * obj.lengths[0]
        else:
            logger.warning(obj.binary_type_enum)
        return results

    def build_class(self, obj: Container) -> Any:
        full_name = obj.class_info.name
        class_name = full_name.split("`1", 1)[0]
        model_class = fullname2classes[class_name]
        alias2key = {
            f.metadata["alias"]: f.name
            for f in dataclasses.fields(model_class)
            if "alias" in f.metadata
        }
        class_kwargs = {}
        for name, value in zip(obj.class_info.member_names, obj.member_values):
            key = alias2key.get(name, name)
            if isinstance(value.value, dict):
                if "real_obj" in value.value:
                    class_kwargs[key] = self.build_object(value.value["real_obj"])
                else:
                    class_kwargs[key] = self.build_object(value.value)
            else:
                class_kwargs[key] = value.value
        if class_name == "System.Collections.Generic.List":
            assert isinstance(class_kwargs["items"], list)
            class_kwargs["items"] = class_kwargs["items"][: class_kwargs["size"]]
        return model_class(**class_kwargs)  # type: ignore[arg-type]

    def build_object(self, obj: Container) -> Any | None:
        if "obj" in obj:
            obj = obj.obj
        if "real_obj" in obj:
            return self.build_object(obj["real_obj"])
        if "Class" in str(obj.record_type_enum):
            return self.build_class(obj)
        elif obj.record_type_enum == RecordTypeEnum.BinaryArray:
            return self.build_binary_array(obj)
        elif obj.record_type_enum == RecordTypeEnum.ArraySinglePrimitive:
            if obj.primitive_type_enum == PrimitiveTypeEnum.Byte:
                return list(obj.member_values)
        elif obj.record_type_enum == RecordTypeEnum.BinaryObjectString:
            return obj.value
        elif "ObjectNullMultiple" in str(obj.record_type_enum):
            return [None] * obj.null_count
        elif obj.record_type_enum not in [
            RecordTypeEnum.ObjectNull,
            RecordTypeEnum.MessageEnd,
        ]:
            logger.warning(obj.record_type_enum)

    def read_record(self, record: Container) -> bool:
        if record.record_type_enum == RecordTypeEnum.SerializedStreamHeader:
            self.header = record.obj
        elif (
            ("Class" in str(record.record_type_enum))
            and (record.obj.class_info.object_id == self.header.root_id)
            and (xstudio_model := self.build_object(record.obj)) is not None
            and isinstance(xstudio_model, XSAppModel)
        ):
            self.xstudio_model = xstudio_model
            return True
        return False

    def resolve_references(self) -> None:
        for ref_id in references_by_id[self.cur_thread_id]:
            ref = references_by_id[self.cur_thread_id][ref_id]
            ref["real_obj"] = self.ref_map[ref["id_ref"]]

    def read(self, path: pathlib.Path) -> tuple[str, XSAppModel]:
        self.svip_file = SVIPFile.parse(path.read_bytes())
        self.resolve_references()

        for record in self.svip_file.record_stream:
            root_found = self.read_record(record)
            if root_found:
                break
        else:
            raise ValueError(_("Root not found"))

        return (
            f"{self.svip_file.magic}{self.svip_file.version}",
            self.xstudio_model,
        )



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import Svip3Project
from .options import InputOptions, OutputOptions
from .svip3_generator import Svip3Generator
from .svip3_parser import Svip3Parser


class Svip3Converter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        svip3_project = Svip3Project.deserialize(path.read_bytes())
        return Svip3Parser(options).parse_project(svip3_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        svip3_project = Svip3Generator(options).generate_project(project)
        path.write_bytes(Svip3Project.serialize(svip3_project))



# mypy: disable-error-code="attr-defined"
import pathlib
from typing import Annotated, get_type_hints

import typer
from rich.progress import track
from rich.prompt import Confirm

from libresvip.cli.prompt import prompt_fields
from libresvip.extension.manager import middleware_manager, plugin_manager
from libresvip.model.base import Project
from libresvip.utils.translation import gettext_lazy as _

app = typer.Typer()


def option_callback(ctx: typer.Context, value: pathlib.Path) -> pathlib.Path | None:
    if ctx.resilient_parsing:
        return None
    ext = value.suffix.lstrip(".").lower()
    if ext not in plugin_manager.plugin_registry:
        raise typer.BadParameter(
            _("Extension {} is not supported. Supported extensions are: {}").format(
                ext, list(plugin_manager.plugin_registry.keys())
            )
        )
    return value


@app.command()
def convert(
    in_path: Annotated[
        pathlib.Path, typer.Argument(exists=True, dir_okay=False, callback=option_callback)
    ],
    out_path: Annotated[
        pathlib.Path, typer.Argument(exists=False, dir_okay=False, callback=option_callback)
    ],
) -> None:
    """
    Convert a file from one format to another.
    """
    input_ext = in_path.suffix.lstrip(".").lower()
    input_plugin = plugin_manager.plugin_registry[input_ext]
    output_ext = out_path.suffix.lstrip(".").lower()
    output_plugin = plugin_manager.plugin_registry[output_ext]
    if (
        input_plugin.plugin_object is not None
        and (input_option := get_type_hints(input_plugin.plugin_object.load).get("options"))
        and output_plugin.plugin_object is not None
        and (output_option := get_type_hints(output_plugin.plugin_object.dump).get("options"))
    ):
        option_type, option_class = _("Input Options: "), input_option
        option_kwargs = {}
        if len(option_class.model_fields):
            typer.echo(option_type)
            option_kwargs = prompt_fields(option_class)
        project = input_plugin.plugin_object.load(in_path, option_class(**option_kwargs))
        for middleware in middleware_manager.plugin_registry.values():
            if (
                (
                    Confirm.ask(
                        _("Enable {} middleware?").format(_(middleware.name)),
                        default=False,
                    )
                )
                and (middleware.plugin_object is not None)
                and (
                    middleware_option := get_type_hints(middleware.plugin_object.process).get(
                        "options"
                    )
                )
            ):
                option_type, option_class = (
                    _("Process Options: "),
                    middleware_option,
                )
                option_kwargs = {}
                if len(option_class.model_fields):
                    typer.echo(option_type)
                    option_kwargs = prompt_fields(option_class)
                project = middleware.plugin_object.process(project, option_class(**option_kwargs))
        option_type, option_class = _("Output Options: "), output_option
        option_kwargs = {}
        if len(option_class.model_fields):
            typer.echo(option_type)
            option_kwargs = prompt_fields(option_class)
        output_plugin.plugin_object.dump(out_path, project, option_class(**option_kwargs))
    else:
        typer.secho("Invalid options", err=True, color=True, fg="red")


@app.command("split")
def split_project(
    in_path: Annotated[
        pathlib.Path, typer.Argument(exists=True, dir_okay=False, callback=option_callback)
    ],
    out_dir: Annotated[pathlib.Path, typer.Argument(exists=True, dir_okay=True)],
    output_ext: Annotated[str, typer.Option(help=_("Output format"))] = "ust",
    max_track_count: Annotated[int, typer.Option(help=_("Maximum track count per file"))] = 1,
) -> None:
    input_ext = in_path.suffix.lstrip(".").lower()
    input_plugin = plugin_manager.plugin_registry[input_ext]
    output_plugin = plugin_manager.plugin_registry[output_ext]
    if (
        input_plugin.plugin_object is not None
        and (input_option := get_type_hints(input_plugin.plugin_object.load).get("options"))
        and output_plugin.plugin_object is not None
        and (output_option := get_type_hints(output_plugin.plugin_object.dump).get("options"))
    ):
        option_type, option_class = _("Input Options: "), input_option
        option_kwargs = {}
        if len(option_class.model_fields):
            typer.echo(option_type)
            option_kwargs = prompt_fields(option_class)
        root_project = input_plugin.plugin_object.load(in_path, option_class(**option_kwargs))
        sub_projects = root_project.split_tracks(max_track_count)
        middleware_with_options = []
        for middleware in middleware_manager.plugin_registry.values():
            if (
                (
                    Confirm.ask(
                        _("Enable {} middleware?").format(_(middleware.name)),
                        default=False,
                    )
                )
                and (middleware.plugin_object is not None)
                and (
                    middleware_option := get_type_hints(middleware.plugin_object.process).get(
                        "options"
                    )
                )
            ):
                option_type, option_class = (
                    _("Process Options: "),
                    middleware_option,
                )
                option_kwargs = {}
                if len(option_class.model_fields):
                    typer.echo(option_type)
                    option_kwargs = prompt_fields(option_class)
                middleware_with_options.append(
                    (
                        middleware.plugin_object.process,
                        option_class,
                        option_kwargs,
                    )
                )
        option_type, option_class = _("Output Options: "), output_option
        option_kwargs = {}
        if len(option_class.model_fields):
            typer.echo(option_type)
            option_kwargs = prompt_fields(option_class)
        for i, project in track(
            enumerate(sub_projects),
            description=_("Converting ..."),
            total=len(sub_projects),
        ):
            for (
                middleware_func,
                option_class,
                option_kwargs,
            ) in middleware_with_options:
                project = middleware_func(project, option_class(**option_kwargs))
            output_plugin.plugin_object.dump(
                out_dir / f"{in_path.stem}_{i + 1:02d}.{output_ext}",
                project,
                option_class(**option_kwargs),
            )
    else:
        typer.secho("Invalid options", err=True, color=True, fg="red")


@app.command("merge")
def merge_projects(
    in_paths: Annotated[list[pathlib.Path], typer.Argument()],
    out_path: Annotated[
        pathlib.Path, typer.Option("", exists=False, dir_okay=False, callback=option_callback)
    ],
) -> None:
    projects = []
    middleware_with_options = []
    for middleware in middleware_manager.plugin_registry.values():
        if (
            (
                Confirm.ask(
                    _("Enable {} middleware?").format(_(middleware.name)),
                    default=False,
                )
            )
            and (middleware.plugin_object is not None)
            and (
                middleware_option := get_type_hints(middleware.plugin_object.process).get("options")
            )
        ):
            option_type, option_class = (
                _("Process Options: "),
                middleware_option,
            )
            option_kwargs = {}
            if len(option_class.model_fields):
                typer.echo(option_type)
                option_kwargs = prompt_fields(option_class)
            middleware_with_options.append(
                (middleware.plugin_object.process, option_class, option_kwargs)
            )
    for in_path in in_paths:
        typer.echo(in_path)
        input_ext = in_path.suffix.lstrip(".").lower()
        input_plugin = plugin_manager.plugin_registry[input_ext]
        if input_plugin.plugin_object is not None and (
            input_option := get_type_hints(input_plugin.plugin_object.load).get("options")
        ):
            option_type, option_class = _("Input Options: "), input_option
            option_kwargs = {}
            if len(option_class.model_fields):
                typer.echo(option_type)
                option_kwargs = prompt_fields(option_class)
            project = input_plugin.plugin_object.load(in_path, option_class(**option_kwargs))
            for (
                middleware_func,
                option_class,
                option_kwargs,
            ) in middleware_with_options:
                project = middleware_func(project, option_class(**option_kwargs))
            projects.append(project)
        else:
            typer.secho("Invalid options", err=True, color=True, fg="red")
            break
    output_ext = out_path.suffix.lstrip(".").lower()
    output_plugin = plugin_manager.plugin_registry[output_ext]
    if (
        projects
        and output_plugin.plugin_object is not None
        and (output_option := get_type_hints(output_plugin.plugin_object.dump).get("options"))
    ):
        project = Project.merge_projects(projects)
        option_type, option_class = _("Output Options: "), output_option
        option_kwargs = {}
        if len(option_class.model_fields):
            typer.echo(option_type)
            option_kwargs = prompt_fields(option_class)
        output_plugin.plugin_object.dump(out_path, project, option_class(**option_kwargs))
    else:
        typer.secho("Invalid options", err=True, color=True, fg="red")



from pydantic import BaseModel, Field

from libresvip.core.constants import DEFAULT_BPM, TICKS_IN_BEAT
from libresvip.model.option_mixins import EnablePitchImportationMixin
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(EnablePitchImportationMixin, BaseModel):
    default_bpm: float = Field(
        default=DEFAULT_BPM,
        title=_("Default BPM"),
        description=_("Used when no BPM information is found in the MIDI file."),
    )


class OutputOptions(BaseModel):
    ticks_per_beat: int = Field(
        default=TICKS_IN_BEAT,
        title=_("Ticks per beat"),
        description=_(
            "Also known as parts per quarter, ticks per quarter note, the number of pulses per quarter note. This setting should not be changed unless you know what it is."
        ),
    )



from typing import Final

LIBRARY_NAME_SINGING_TOOL_MODEL: Final[str] = (
    "SingingTool.Model, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null"
)
LIBRARY_NAME_SINGING_TOOL_LIBRARY: Final[str] = (
    "SingingTool.Library, Version=1.0.0.0, Culture=neutral, PublicKeyToken=null"
)

# Values
VALUE_LIST_VERSION_SONG_TEMPO: Final[int] = 2
VALUE_LIST_VERSION_SONG_BEAT: Final[int] = 2
VALUE_LIST_VERSION_SONG_ITRACK: Final[int] = 4
VALUE_LIST_VERSION_SONG_NOTE: Final[int] = 1054



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import VxFile
from .options import InputOptions, OutputOptions
from .vx_beta_generator import VxBetaGenerator
from .vx_beta_parser import VxBetaParser


class VxBetaConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        vx_file = VxFile.parse(path.read_bytes())
        return VxBetaParser(options).parse_project(vx_file)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        vx_file = VxBetaGenerator(options).generate_project(project)
        path.write_bytes(VxFile.build(vx_file))



from libresvip.extension import base as plugin_base
from libresvip.model.base import ParamCurve, Points, Project, SingingTrack

from .options import ProcessOptions


class PitchShiftMiddleware(plugin_base.MiddlewareBase):
    def process(self, project: Project, options: ProcessOptions) -> Project:
        if options.key:
            return project.model_copy(
                update={
                    "track_list": [
                        track.model_copy(
                            update={
                                "note_list": [
                                    note.model_copy(
                                        update={"key_number": note.key_number + options.key}
                                    )
                                    for note in track.note_list
                                ],
                                "edited_params": track.edited_params.model_copy(
                                    update={
                                        "pitch": ParamCurve(
                                            points=Points(
                                                root=[
                                                    point
                                                    if point.y == -100
                                                    else point._replace(
                                                        y=point.y + options.key * 100
                                                    )
                                                    for point in track.edited_params.pitch.points.root
                                                ]
                                            )
                                        )
                                    }
                                ),
                            }
                        )
                        if isinstance(track, SingingTrack)
                        else track
                        for track in project.track_list
                    ]
                }
            )
        else:
            return project



import abc

from xsdata_pydantic.fields import field

from libresvip.model.base import BaseModel
from libresvip.utils.music_math import note2midi


class VocalSharpParamBase(abc.ABC, BaseModel):
    time: int = field(
        metadata={
            "name": "t",
            "type": "Element",
        },
    )
    value: int = field(
        metadata={
            "name": "v",
            "type": "Element",
        },
    )


class BRE(VocalSharpParamBase):
    class Meta:
        name = "B"


class DYN(VocalSharpParamBase):
    class Meta:
        name = "D"


class GEN(VocalSharpParamBase):
    class Meta:
        name = "G"


class PIT(VocalSharpParamBase):
    class Meta:
        name = "P"


class BRI(VocalSharpParamBase):
    class Meta:
        name = "R"


class STR(VocalSharpParamBase):
    class Meta:
        name = "S"


class VOC(VocalSharpParamBase):
    class Meta:
        name = "V"


class GWL(VocalSharpParamBase):
    class Meta:
        name = "W"


class XSY(VocalSharpParamBase):
    class Meta:
        name = "X"


class VocalSharpBeat(BaseModel):
    class Meta:
        name = "Beat"

    bar_index: int | None = field(
        default=0,
        metadata={
            "name": "barIndex",
            "type": "Element",
        },
    )
    beat_per_bar: int | None = field(
        default=4,
        metadata={
            "name": "beatPerBar",
            "type": "Element",
        },
    )
    bar_divide: int | None = field(
        default=4,
        metadata={
            "name": "barDivide",
            "type": "Element",
        },
    )


class VocalSharpDefaultParameter(BaseModel):
    class Meta:
        name = "DefaultParameter"

    dyn: int | None = field(
        default=128,
        metadata={
            "type": "Element",
        },
    )
    strength: int | None = field(
        default=128,
        metadata={
            "name": "str",
            "type": "Element",
        },
    )
    bri: int | None = field(
        default=128,
        metadata={
            "type": "Element",
        },
    )
    voc: int | None = field(
        default=255,
        metadata={
            "type": "Element",
        },
    )
    bre: int | None = field(
        default=128,
        metadata={
            "type": "Element",
        },
    )
    gen: int | None = field(
        default=128,
        metadata={
            "type": "Element",
        },
    )
    gwl: int | None = field(
        default=0,
        metadata={
            "type": "Element",
        },
    )
    xsy: int | None = field(
        default=0,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpTrillBase(abc.ABC, BaseModel):
    pos: float = field(
        default=0.25,
        metadata={
            "type": "Element",
        },
    )
    amplitude: float = field(
        default=0,
        metadata={
            "type": "Element",
        },
    )
    frequency: float = field(
        default=5.5,
        metadata={
            "type": "Element",
        },
    )
    phase: float = field(
        default=0,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpTrill(VocalSharpTrillBase):
    class Meta:
        name = "trill"


class VocalSharpDefaultTrill(VocalSharpTrillBase):
    class Meta:
        name = "DefaultTrill"


class VocalSharpSequence(BaseModel):
    class Meta:
        name = "Sequence"

    pos: int | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    name: str | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    path: str | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpTrackBase(abc.ABC, BaseModel):
    name: str | None = field(
        default="神秘轨道_0",
        metadata={
            "name": "Name",
            "type": "Element",
        },
    )
    pan: float | None = field(
        default=0,
        metadata={
            "name": "Pan",
            "type": "Element",
        },
    )
    gain: float | None = field(
        default=0,
        metadata={
            "name": "Gain",
            "type": "Element",
        },
    )
    is_mute: str | None = field(
        default="False",
        metadata={
            "name": "IsMute",
            "type": "Element",
        },
    )
    is_solo: str | None = field(
        default="False",
        metadata={
            "name": "IsSolo",
            "type": "Element",
        },
    )


class VocalSharpInstrumentalTrackBase(VocalSharpTrackBase, abc.ABC):
    sequences: list[VocalSharpSequence] = field(
        default_factory=list,
        metadata={
            "name": "Sequence",
            "type": "Element",
        },
    )


class VocalSharpStereoTrack(VocalSharpInstrumentalTrackBase):
    class Meta:
        name = "StereoTrack"


class VocalSharpMonoTrack(VocalSharpInstrumentalTrackBase):
    class Meta:
        name = "MonoTrack"


class VocalSharpTempo(BaseModel):
    class Meta:
        name = "Tempo"

    pos: int | None = field(
        default=0,
        metadata={
            "type": "Element",
        },
    )
    bpm: float | None = field(
        default=120,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpPoint(BaseModel):
    class Meta:
        name = "p"

    x: float | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    y: float | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpParameter(BaseModel):
    class Meta:
        name = "Parameter"

    points: list[object] = field(
        default_factory=list,
        metadata={
            "type": "Wildcard",
            "mixed": True,
            "choices": (
                {"name": "B", "type": BRE},
                {"name": "D", "type": DYN},
                {"name": "G", "type": GEN},
                {"name": "P", "type": PIT},
                {"name": "R", "type": BRI},
                {"name": "S", "type": STR},
                {"name": "V", "type": VOC},
                {"name": "W", "type": GWL},
                {"name": "X", "type": XSY},
            ),
        },
    )


class VocalSharpSyllablePartBase(abc.ABC, BaseModel):
    p: list[VocalSharpPoint] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        },
    )
    sample: str | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    symbol: str | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpSyllableHead(VocalSharpSyllablePartBase):
    class Meta:
        name = "head"


class VocalSharpSyllableCur(VocalSharpSyllablePartBase):
    class Meta:
        name = "cur"


class VocalSharpSyllableTail(VocalSharpSyllablePartBase):
    class Meta:
        name = "tail"


class VocalSharpSyllable(BaseModel):
    class Meta:
        name = "Syllable"

    head: VocalSharpSyllableHead | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    cur: VocalSharpSyllableCur | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    tail: VocalSharpSyllableTail | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )


class VocalSharpNote(BaseModel):
    class Meta:
        name = "Note"

    pitch: str = field(
        metadata={
            "type": "Element",
        },
    )
    pos: int = field(
        metadata={
            "type": "Element",
        },
    )
    duration: int = field(
        metadata={
            "type": "Element",
        },
    )
    lyric: str | None = field(
        default=None,
        metadata={
            "type": "Element",
        },
    )
    trill: VocalSharpTrill | None = field(
        default=None,
        metadata={
            "name": "trill",
            "type": "Element",
        },
    )
    syllable: list[VocalSharpSyllable] = field(
        default_factory=list,
        metadata={
            "name": "Syllable",
            "type": "Element",
        },
    )

    @property
    def key_number(self) -> int:
        return note2midi(self.pitch)


class VocalSharpNoteTrack(VocalSharpTrackBase):
    class Meta:
        name = "NoteTrack"

    singer: str | None = field(
        default="神秘歌手",
        metadata={
            "name": "Singer",
            "type": "Element",
        },
    )
    lsd: str | None = field(
        default="神秘字典",
        metadata={
            "name": "LSD",
            "type": "Element",
        },
    )
    default_parameter: VocalSharpDefaultParameter | None = field(
        default_factory=VocalSharpDefaultParameter,
        metadata={
            "name": "DefaultParameter",
            "type": "Element",
        },
    )
    por: float = field(
        default=0.07,
        metadata={
            "type": "Element",
        },
    )
    note: list[VocalSharpNote] = field(
        default_factory=list,
        metadata={
            "name": "Note",
            "type": "Element",
        },
    )
    parameter: VocalSharpParameter = field(
        default_factory=VocalSharpParameter,
        metadata={
            "name": "Parameter",
            "type": "Element",
        },
    )


class VocalSharpInnerProject(BaseModel):
    class Meta:
        name = "Project"

    samples_per_sec: int | None = field(
        default=44100,
        metadata={
            "name": "SamplesPerSec",
            "type": "Element",
        },
    )
    resolution: int | None = field(
        default=1920,
        metadata={
            "name": "Resolution",
            "type": "Element",
        },
    )
    duration: int | None = field(
        default=None,
        metadata={
            "name": "Duration",
            "type": "Element",
        },
    )
    default_trill: VocalSharpDefaultTrill | None = field(
        default_factory=VocalSharpDefaultTrill,
        metadata={
            "name": "DefaultTrill",
            "type": "Element",
        },
    )
    tempo: list[VocalSharpTempo] = field(
        default_factory=list,
        metadata={
            "name": "Tempo",
            "type": "Element",
        },
    )
    beat: list[VocalSharpBeat] = field(
        default_factory=list,
        metadata={
            "name": "Beat",
            "type": "Element",
        },
    )
    tracks: list[object] = field(
        default_factory=list,
        metadata={
            "type": "Elements",
            "choices": (
                {"name": "NoteTrack", "type": VocalSharpNoteTrack},
                {"name": "StereoTrack", "type": VocalSharpStereoTrack},
                {"name": "MonoTrack", "type": VocalSharpMonoTrack},
            ),
        },
    )


class VocalSharpProject(BaseModel):
    class Meta:
        name = "VSPX"

    version: str | None = field(
        default="Beta0.0.0",
        metadata={
            "name": "Version",
            "type": "Element",
        },
    )
    project: VocalSharpInnerProject = field(
        default_factory=VocalSharpInnerProject,
        metadata={
            "name": "Project",
            "type": "Element",
        },
    )



import math
from dataclasses import dataclass
from typing import TYPE_CHECKING

from libresvip.core.constants import (
    DEFAULT_PITCH_BEND_SENSITIVITY,
    MAX_PITCH_BEND_SENSITIVITY,
    MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS,
    PITCH_MAX_VALUE,
)
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.music_math import clamp

if TYPE_CHECKING:
    from libresvip.model.point import Point


@dataclass
class ControlEvent:
    tick: int
    value: int


@dataclass
class MIDIPitchData:
    pit: list[ControlEvent]
    pbs: list[ControlEvent]


def generate_for_midi(
    first_bar_length: int,
    pitch: ParamCurve,
    notes: list[Note],
    synchronizer: TimeSynchronizer,
) -> MIDIPitchData | None:
    pitch_simulator = PitchSimulator(
        synchronizer=synchronizer,
        portamento=PortamentoPitch.no_portamento(),
        note_list=notes,
    )
    data = RelativePitchCurve(first_bar_length).from_absolute(pitch.points.root, pitch_simulator)
    if not len(data):
        return None
    pitch_sectioned: list[list[Point]] = [[]]
    current_pos = 0
    for pitch_event in data:
        if (
            not pitch_sectioned[-1]
            or pitch_event.x - current_pos < MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS
        ):
            pitch_sectioned[-1].append(pitch_event)
        else:
            pitch_sectioned.append([pitch_event])
        current_pos = pitch_event.x
    pit: list[ControlEvent] = []
    pbs: list[ControlEvent] = []
    for section in pitch_sectioned:
        if len(section):
            max_abs_value = max(abs(point.y / 100) for point in section)
            pbs_for_this_section = min(math.ceil(max_abs_value), MAX_PITCH_BEND_SENSITIVITY)
            if pbs_for_this_section > DEFAULT_PITCH_BEND_SENSITIVITY:
                pbs.extend(
                    (
                        ControlEvent(
                            section[-1][0] + MIN_BREAK_LENGTH_BETWEEN_PITCH_SECTIONS // 2,
                            DEFAULT_PITCH_BEND_SENSITIVITY,
                        ),
                    )
                )
            else:
                pbs_for_this_section = DEFAULT_PITCH_BEND_SENSITIVITY
            pit.extend(
                ControlEvent(
                    pitch_pos,
                    int(
                        clamp(
                            pitch_value
                            * (PITCH_MAX_VALUE if pitch_value > 0 else (PITCH_MAX_VALUE + 1))
                            / 100
                            / pbs_for_this_section,
                            -PITCH_MAX_VALUE - 1,
                            PITCH_MAX_VALUE,
                        )
                    ),
                )
                for pitch_pos, pitch_value in section
            )
    return MIDIPitchData(pit=pit, pbs=pbs)



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.model.option_mixins import SelectSingleTrackMixin
from libresvip.utils.translation import gettext_lazy as _


class SplitOption(Enum):
    BOTH: Annotated[
        str,
        Field(
            title=_("Both note gap and punctuation"),
            description=_(
                "When the interval between two adjacent notes is greater than or equal to thirty-second note or a punctuation mark is encountered, start a new line."
            ),
        ),
    ] = "both"
    GAP: Annotated[
        str,
        Field(
            title=_("Note gap only"),
            description=_(
                "When the interval between two adjacent notes is greater than or equal to thirty-second note, start a new line."
            ),
        ),
    ] = "gap"
    SYMBOL: Annotated[
        str,
        Field(
            title=_("Punctuation only"),
            description=_("When a punctuation mark is encountered, start a new line."),
        ),
    ] = "symbol"


class OutputOptions(SelectSingleTrackMixin, BaseModel):
    offset: int = Field(
        0,
        title=_("Offset"),
        description=_("In milliseconds, positive means ahead, negative means opposite."),
    )
    split_by: SplitOption = Field(title=_("New line by"), default=SplitOption.BOTH)
    ignore_slur_notes: bool = Field(
        title=_("Ignore slur notes"),
        description=_(
            "Ignore '-' lyrics that are used to indicate slur notes in singing synthesizers."
        ),
        default=True,
    )
    encoding: str = Field(title=_("Text encoding"), default="utf-8")



from functools import partial

from libresvip.model.base import BaseModel


class TimeTag(BaseModel):
    minute: int
    second: int
    milisecond: int


class LyricLine(BaseModel):
    time_tags: list[TimeTag]
    lyric: str | None = None


class InfoTag(BaseModel):
    key: str
    value: str


class LrcFile(BaseModel):
    info_tags: list[InfoTag]
    lyric_lines: list[LyricLine]


TitleInfoTag = partial(InfoTag, key="ti")
ArtistInfoTag = partial(InfoTag, key="ar")
AlbumInfoTag = partial(InfoTag, key="al")
ByInfoTag = partial(InfoTag, key="by")
OffsetInfoTag = partial(InfoTag, key="offset")



from __future__ import annotations

import contextlib
import dataclasses
import enum
import pathlib
import traceback
import zipfile
from typing import Any, get_args, get_type_hints

import more_itertools
from loguru import logger
from pydantic_core import PydanticUndefined
from pydantic_extra_types.color import Color
from PySide6.QtCore import (
    Property,
    QAbstractListModel,
    QModelIndex,
    QObject,
    QRunnable,
    Qt,
    QThreadPool,
    QTimer,
    Signal,
    Slot,
)
from PySide6.QtQml import QmlElement
from upath import UPath

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401

from libresvip.core.config import ConversionMode, get_ui_settings, settings
from libresvip.core.warning_types import CatchWarnings
from libresvip.extension.manager import middleware_manager, plugin_manager
from libresvip.gui.models.base_task import BaseTask
from libresvip.gui.models.list_models import ModelProxy
from libresvip.gui.models.table_models import PluginCadidatesTableModel
from libresvip.model.base import BaseComplexModel, BaseModel, Project
from libresvip.utils.text import supported_charset_names, uuid_str

from .url_opener import open_path

QML_IMPORT_NAME = "LibreSVIP"
QML_IMPORT_MAJOR_VERSION = 1
QML_IMPORT_MINOR_VERSION = 0


class ConversionWorkerSignals(QObject):
    result = Signal(int, dict)


class ConversionWorker(QRunnable):
    def __init__(
        self,
        index: int,
        input_path: str,
        output_path: str,
        input_format: str,
        output_format: str,
        input_options: dict[str, Any],
        output_options: dict[str, Any],
        middleware_options: dict[str, dict[str, Any]],
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.index = index
        self.input_path = input_path
        self.output_path = output_path
        self.input_format = input_format
        self.output_format = output_format
        self.input_options = input_options
        self.output_options = output_options
        self.middleware_options = middleware_options
        self.signals = ConversionWorkerSignals()

    @Slot()
    def run(self) -> None:
        result_kwargs: dict[str, Any] = {"running": False}
        try:
            with CatchWarnings() as w:
                input_plugin = plugin_manager.plugin_registry[self.input_format]
                output_plugin = plugin_manager.plugin_registry[self.output_format]
                if (
                    input_plugin.plugin_object is None
                    or (
                        input_option_cls := get_type_hints(input_plugin.plugin_object.load).get(
                            "options",
                        )
                    )
                    is None
                    or output_plugin.plugin_object is None
                    or (
                        output_option_cls := get_type_hints(
                            output_plugin.plugin_object.dump,
                        ).get("options")
                    )
                    is None
                ):
                    result_kwargs |= {
                        "success": False,
                        "error": "Invalid options",
                    }
                else:
                    project = input_plugin.plugin_object.load(
                        pathlib.Path(self.input_path),
                        input_option_cls(**self.input_options),
                    )
                    for (
                        middleware_abbr,
                        middleware_option,
                    ) in self.middleware_options.items():
                        middleware = middleware_manager.plugin_registry[middleware_abbr]
                        if (
                            middleware.plugin_object is not None
                            and hasattr(middleware.plugin_object, "process")
                            and (
                                middleware_option_class := get_type_hints(
                                    middleware.plugin_object.process
                                ).get(
                                    "options",
                                )
                            )
                        ):
                            project = middleware.plugin_object.process(
                                project,
                                middleware_option_class.model_validate(middleware_option),
                            )
                    output_plugin.plugin_object.dump(
                        UPath(self.output_path),
                        project,
                        output_option_cls(**self.output_options),
                    )
                    result_kwargs |= {
                        "success": True,
                        "tmp_path": self.output_path,
                    }
            if w.output:
                result_kwargs["warning"] = w.output
            self.signals.result.emit(self.index, result_kwargs)
        except Exception:
            result_kwargs |= {
                "success": False,
                "error": traceback.format_exc(),
            }
            with contextlib.suppress(RuntimeError):
                self.signals.result.emit(self.index, result_kwargs)


class SplitWorker(QRunnable):
    def __init__(
        self,
        index: int,
        input_path: str,
        output_dir: str,
        input_format: str,
        output_format: str,
        input_options: dict[str, Any],
        output_options: dict[str, Any],
        middleware_options: dict[str, dict[str, Any]],
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.index = index
        self.input_path = input_path
        self.output_dir = output_dir
        UPath(self.output_dir).mkdir(parents=True, exist_ok=True)
        self.input_format = input_format
        self.output_format = output_format
        self.input_options = input_options
        self.output_options = output_options
        self.middleware_options = middleware_options
        self.signals = ConversionWorkerSignals()

    @Slot()
    def run(self) -> None:
        result_kwargs: dict[str, Any] = {"running": False}
        try:
            with CatchWarnings() as w:
                input_plugin = plugin_manager.plugin_registry[self.input_format]
                output_plugin = plugin_manager.plugin_registry[self.output_format]
                if (
                    input_plugin.plugin_object is None
                    or (
                        input_option_cls := get_type_hints(input_plugin.plugin_object.load).get(
                            "options",
                        )
                    )
                    is None
                    or output_plugin.plugin_object is None
                    or (
                        output_option_cls := get_type_hints(
                            output_plugin.plugin_object.dump,
                        ).get("options")
                    )
                    is None
                ):
                    result_kwargs |= {
                        "success": False,
                        "error": "Invalid options",
                    }
                else:
                    project = input_plugin.plugin_object.load(
                        pathlib.Path(self.input_path),
                        input_option_cls(**self.input_options),
                    )
                    for (
                        middleware_abbr,
                        middleware_option,
                    ) in self.middleware_options.items():
                        middleware = middleware_manager.plugin_registry[middleware_abbr]
                        if (
                            middleware.plugin_object is not None
                            and hasattr(middleware.plugin_object, "process")
                            and (
                                middleware_option_class := get_type_hints(
                                    middleware.plugin_object.process
                                ).get(
                                    "options",
                                )
                            )
                        ):
                            project = middleware.plugin_object.process(
                                project,
                                middleware_option_class.model_validate(middleware_option),
                            )
                    output_option = output_option_cls(**self.output_options)
                    for child_project in project.split_tracks(settings.max_track_count):
                        output_plugin.plugin_object.dump(
                            UPath(self.output_dir) / uuid_str(),
                            child_project,
                            output_option,
                        )
                    result_kwargs |= {
                        "success": True,
                        "tmp_path": self.output_dir,
                    }
            if w.output:
                result_kwargs["warning"] = w.output
            self.signals.result.emit(self.index, result_kwargs)
        except Exception:
            result_kwargs |= {
                "success": False,
                "error": traceback.format_exc(),
            }
            with contextlib.suppress(RuntimeError):
                self.signals.result.emit(self.index, result_kwargs)


class MergeWorker(QRunnable):
    def __init__(
        self,
        index: int,
        input_paths: list[str],
        output_path: str,
        input_format: str,
        output_format: str,
        input_options: dict[str, Any],
        output_options: dict[str, Any],
        middleware_options: dict[str, dict[str, Any]],
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.index = index
        self.input_paths = input_paths
        self.output_path = output_path
        self.input_format = input_format
        self.output_format = output_format
        self.input_options = input_options
        self.output_options = output_options
        self.middleware_options = middleware_options
        self.signals = ConversionWorkerSignals()

    @Slot()
    def run(self) -> None:
        result_kwargs: dict[str, Any] = {"running": False}
        try:
            with CatchWarnings() as w:
                input_plugin = plugin_manager.plugin_registry[self.input_format]
                output_plugin = plugin_manager.plugin_registry[self.output_format]
                if (
                    input_plugin.plugin_object is None
                    or (
                        input_option_cls := get_type_hints(input_plugin.plugin_object.load).get(
                            "options",
                        )
                    )
                    is None
                    or output_plugin.plugin_object is None
                    or (
                        output_option_cls := get_type_hints(
                            output_plugin.plugin_object.dump,
                        ).get("options")
                    )
                    is None
                ):
                    result_kwargs |= {
                        "success": False,
                        "error": "Invalid options",
                    }
                else:
                    input_option = input_option_cls(**self.input_options)
                    child_projects = [
                        input_plugin.plugin_object.load(
                            pathlib.Path(input_path),
                            input_option,
                        )
                        for input_path in self.input_paths
                    ]
                    project = Project.merge_projects(child_projects)
                    for (
                        middleware_abbr,
                        middleware_option,
                    ) in self.middleware_options.items():
                        middleware = middleware_manager.plugin_registry[middleware_abbr]
                        if (
                            middleware.plugin_object is not None
                            and hasattr(middleware.plugin_object, "process")
                            and (
                                middleware_option_class := get_type_hints(
                                    middleware.plugin_object.process
                                ).get(
                                    "options",
                                )
                            )
                        ):
                            project = middleware.plugin_object.process(
                                project,
                                middleware_option_class.model_validate(middleware_option),
                            )
                    output_plugin.plugin_object.dump(
                        UPath(self.output_path),
                        project,
                        output_option_cls(**self.output_options),
                    )
                    result_kwargs |= {
                        "success": True,
                        "tmp_path": self.output_path,
                    }
            if w.output:
                result_kwargs["warning"] = w.output
            self.signals.result.emit(self.index, result_kwargs)
        except Exception:
            result_kwargs |= {
                "success": False,
                "error": traceback.format_exc(),
            }
            with contextlib.suppress(RuntimeError):
                self.signals.result.emit(self.index, result_kwargs)


@QmlElement
class TaskManager(QObject):
    conversion_mode_changed = Signal(str)
    input_format_changed = Signal(str)
    output_format_changed = Signal(str)
    input_fields_changed = Signal()
    output_fields_changed = Signal()
    task_count_changed = Signal(int)
    busy_changed = Signal(bool)
    middleware_options_updated = Signal()
    _start_conversion = Signal()

    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent=parent)
        self._conversion_mode = ConversionMode.DIRECT
        self.tasks = ModelProxy(dataclasses.asdict(BaseTask()))
        self.tasks.rowsInserted.connect(self._on_tasks_changed)
        self.tasks.rowsRemoved.connect(self._on_tasks_changed)
        self.input_formats = ModelProxy({"value": "", "text": ""})
        self.output_formats = ModelProxy({"value": "", "text": ""})
        self.input_fields = ModelProxy(
            {
                "index": 0,
                "name": "",
                "title": "",
                "description": "",
                "default": "",
                "type": "",
                "value": "",
                "choices": [],
            }
        )
        self._input_fields_inited = False
        self.output_fields = ModelProxy(
            {
                "index": 0,
                "name": "",
                "title": "",
                "description": "",
                "default": "",
                "type": "",
                "value": "",
                "choices": [],
            }
        )
        self._output_fields_inited = False
        self.middleware_states = ModelProxy(
            {
                "index": 0,
                "identifier": "",
                "name": "",
                "description": "",
                "value": False,
            }
        )
        self.middleware_fields: dict[str, ModelProxy] = {}
        self.init_middleware_options()
        self.reload_formats()
        self.thread_pool = QThreadPool.global_instance()
        self.input_format_changed.connect(self.set_input_fields)
        self.output_format_changed.connect(self.set_output_fields)
        self.output_format_changed.connect(self.reset_output_ext)
        self._start_conversion.connect(self.start)
        self.timer = QTimer()
        self.timer.interval = 100
        self.timer.timeout.connect(self.check_busy)
        self.tasks.rowsAboutToBeRemoved.connect(self.delete_tmp_file)
        self.plugin_candidates = PluginCadidatesTableModel()

    @Slot(int)
    def toggle_plugin(self, index: int) -> None:
        key = plugin_manager._candidates[index][1].suffix
        if key in plugin_manager.plugin_registry and key not in settings.disabled_plugins:
            settings.disabled_plugins.append(key)
        elif key in settings.disabled_plugins:
            settings.disabled_plugins.remove(key)
        else:
            return
        plugin_manager.import_plugins(reload=True)
        self.plugin_candidates.reload_formats()
        self.reload_formats()

    def _on_tasks_changed(self, index: QModelIndex, start: int, end: int) -> None:
        self.task_count_changed.emit(len(self.tasks))

    def get_conversion_mode(self) -> str:
        return self._conversion_mode.value

    def set_conversion_mode(self, mode: str) -> None:
        self._conversion_mode = ConversionMode(mode)
        self.conversion_mode_changed.emit(mode)
        for i, task in enumerate(self.tasks):
            if task["success"] is False or self.delete_tmp_file(QModelIndex(), i, i):
                self.tasks.update(
                    i,
                    {
                        "tmp_path": "",
                        "running": False,
                        "success": None,
                        "error": "",
                        "warning": "",
                    },
                )

    conversion_mode = Property(
        str,
        get_conversion_mode,
        set_conversion_mode,
        notify=conversion_mode_changed,
    )

    @Property(int, notify=task_count_changed)
    def count(self) -> int:
        return len(self.tasks)

    def init_middleware_options(self) -> None:
        for middleware in middleware_manager.plugin_registry.values():
            if middleware.plugin_object is not None and hasattr(
                middleware.plugin_object, "process"
            ):
                self.middleware_states.append(
                    {
                        "index": len(self.middleware_states),
                        "identifier": middleware.identifier,
                        "name": middleware.name,
                        "description": middleware.description,
                        "value": False,
                    }
                )
                self.middleware_fields[middleware.identifier] = ModelProxy(
                    {
                        "name": "",
                        "title": "",
                        "description": "",
                        "default": "",
                        "type": "",
                        "value": "",
                        "choices": [],
                    }
                )
        self.reload_middleware_options()
        self.middleware_options_updated.connect(self.reload_middleware_options)

    def reload_middleware_options(self) -> None:
        for middleware in middleware_manager.plugin_registry.values():
            if middleware.plugin_object is not None and hasattr(
                middleware.plugin_object, "process"
            ):
                option_class = get_type_hints(middleware.plugin_object.process)["options"]
                self.middleware_fields[middleware.identifier].clear()
                middleware_fields = self.inspect_fields(option_class)
                self.middleware_fields[middleware.identifier].append_many(middleware_fields)

    @Slot(result=None)
    def reload_formats(self) -> None:
        self.input_formats.clear()
        self.input_formats.append_many(
            [
                {
                    "text": plugin.file_format,
                    "value": plugin.suffix,
                }
                for plugin in plugin_manager.plugin_registry.values()
            ],
        )
        self.input_format_changed.emit("")
        self.output_formats.clear()
        self.output_formats.append_many(
            [
                {
                    "text": plugin.file_format,
                    "value": plugin.suffix,
                }
                for plugin in plugin_manager.plugin_registry.values()
            ],
        )
        self.output_format_changed.emit("")

    def delete_tmp_file(self, index: QModelIndex, start: int, end: int) -> bool:
        deleted = False
        for i in range(start, end + 1):
            path = UPath(self.tasks[i]["tmp_path"])
            if path.exists():
                with contextlib.suppress(Exception):
                    if path.is_dir():
                        path.rmdir()
                    else:
                        path.unlink()
                deleted = True
        return deleted

    @property
    def input_format(self) -> str | None:
        return settings.last_input_format

    @input_format.setter
    def input_format(self, value: str | None) -> None:
        if value is not None:
            settings.last_input_format = value

    @property
    def output_format(self) -> str | None:
        return settings.last_output_format

    @output_format.setter
    def output_format(self, value: str | None) -> None:
        if value is not None:
            settings.last_output_format = value

    @Slot(result=bool)
    def start_conversion(self) -> bool:
        if self.busy:
            return False
        self._start_conversion.emit()
        return True

    @property
    def output_ext(self) -> str:
        if settings.auto_set_output_extension and self.output_format:
            return f".{self.output_format}"
        return ""

    @Slot(bool)
    def reset_output_ext(self, value: bool) -> None:
        self.tasks.update_many(0, [{"ext": self.output_ext}] * len(self.tasks))

    @staticmethod
    def output_dir(task: dict[str, Any]) -> pathlib.Path:
        return (
            settings.save_folder
            if settings.save_folder.is_absolute()
            else pathlib.Path(task["path"]).parent / str(settings.save_folder)
        )

    def output_path(self, task: dict[str, Any]) -> pathlib.Path:
        output_dir = self.output_dir(task)
        if self._conversion_mode == ConversionMode.SPLIT:
            return output_dir / f"{task['stem']}_**{self.output_ext}"
        return output_dir / f"{task['stem']}{self.output_ext}"

    @Slot(int, result=str)
    def get_output_path(self, index: int) -> str:
        task = self.tasks[index]
        return str(self.output_path(task))

    @Slot(int, result=bool)
    def open_output_dir(self, index: int) -> bool:
        task = self.tasks[index]
        open_path(self.output_dir(task))
        return True

    @Slot(int, result=bool)
    def open_output_path(self, index: int) -> bool:
        task = self.tasks[index]
        open_path(self.output_path(task))
        return True

    @Slot(int, result=bool)
    def output_path_exists(self, index: int) -> bool:
        task = self.tasks[index]
        if self._conversion_mode != ConversionMode.SPLIT:
            return self.output_path(task).exists()
        output_dir = self.output_dir(task)
        tmp_path = UPath(task["tmp_path"])
        return any(
            target_path.exists()
            for i in range(more_itertools.ilen(tmp_path.iterdir()))
            if (target_path := output_dir / f"{task['stem']}_{i + 1:0=2d}{self.output_ext}")
        )

    @Slot(int, result=bool)
    def move_to_output(self, index: int) -> bool:
        task = self.tasks[index]
        if not task["success"]:
            return False
        output_dir = self.output_dir(task)
        tmp_path = UPath(task["tmp_path"])
        result = False
        try:
            if tmp_path.exists():
                if tmp_path.is_dir():
                    for i, child_file in enumerate(tmp_path.iterdir()):
                        if not child_file.is_file():
                            continue
                        target_path = output_dir / f"{task['stem']}_{i + 1:0=2d}{self.output_ext}"
                        if target_path.exists():
                            target_path.unlink()
                        target_path.write_bytes(child_file.read_bytes())
                else:
                    target_path = output_dir / f"{task['stem']}{self.output_ext}"
                    if target_path.exists():
                        target_path.unlink()
                    target_path.write_bytes(tmp_path.read_bytes())
                result = True
        except OSError as e:
            self.tasks.update(index, {"error": str(e)})
        return result

    @staticmethod
    def inspect_fields(option_class: BaseModel) -> list[dict[str, Any]]:
        fields = []
        for i, (option_key, field_info) in enumerate(option_class.model_fields.items()):
            default_value = None if field_info.default is PydanticUndefined else field_info.default
            if issubclass(field_info.annotation, bool):
                fields.append(
                    {
                        "index": i,
                        "type": "bool",
                        "name": option_key,
                        "title": field_info.title,
                        "description": field_info.description or "",
                        "default": default_value,
                        "value": default_value,
                    }
                )
            elif issubclass(field_info.annotation, enum.Enum):
                if default_value is not None:
                    default_value = default_value.value
                annotations = get_type_hints(field_info.annotation, include_extras=True)
                choices = []
                for enum_item in field_info.annotation:
                    if enum_item.name in annotations:
                        annotated_args = list(get_args(annotations[enum_item.name]))
                        if len(annotated_args) >= 2:
                            _, enum_field = annotated_args[:2]
                        else:
                            continue
                        choices.append(
                            {
                                "value": enum_item.value,
                                "text": enum_field.title,
                                "desc": enum_field.description or "",
                            }
                        )
                    else:
                        logger.warning(enum_item.name)
                fields.append(
                    {
                        "index": i,
                        "type": "enum",
                        "name": option_key,
                        "title": field_info.title,
                        "description": field_info.description or "",
                        "default": default_value,
                        "value": default_value,
                        "choices": choices,
                    }
                )
            elif option_key == "lyric_replacement_preset_name":
                choices = [
                    {
                        "value": preset,
                        "text": preset,
                        "desc": "",
                    }
                    for preset in get_ui_settings().lyric_replace_rules
                ]
                fields.append(
                    {
                        "index": i,
                        "type": "enum",
                        "name": option_key,
                        "title": field_info.title,
                        "description": field_info.description or "",
                        "default": default_value,
                        "value": default_value,
                        "choices": choices,
                    }
                )
            elif option_key in ["encoding", "lyric_encoding"]:
                choices = [
                    {
                        "value": preset,
                        "text": preset,
                        "desc": "",
                    }
                    for preset in supported_charset_names()
                ]
                fields.append(
                    {
                        "index": i,
                        "type": "enum",
                        "name": option_key,
                        "title": field_info.title,
                        "description": field_info.description or "",
                        "default": default_value,
                        "value": default_value,
                        "choices": choices,
                    }
                )
            elif issubclass(
                field_info.annotation,
                str | int | float | Color | BaseComplexModel,
            ):
                if issubclass(field_info.annotation, BaseComplexModel):
                    default_value = field_info.annotation.default_repr()
                fields.append(
                    {
                        "index": i,
                        "type": "color"
                        if issubclass(field_info.annotation, Color)
                        else field_info.annotation.__name__,
                        "name": option_key,
                        "title": field_info.title,
                        "description": field_info.description or "",
                        "default": default_value,
                        "value": default_value,
                    }
                )
        return fields

    @Slot(str)
    def set_input_fields(self, input_format: str) -> None:
        if input_format == self.input_format and self._input_fields_inited:
            return
        if input_format:
            self.input_format = input_format
            self.input_fields.clear()
            plugin_input = plugin_manager.plugin_registry[self.input_format]
            if plugin_input.plugin_object is not None and hasattr(
                plugin_input.plugin_object, "load"
            ):
                option_class = get_type_hints(plugin_input.plugin_object.load)["options"]
                input_fields = self.inspect_fields(option_class)
                self.input_fields.append_many(input_fields)
            if not self._input_fields_inited:
                self._input_fields_inited = True
            self.input_fields_changed.emit()

    @Slot(str)
    def set_output_fields(self, output_format: str) -> None:
        if output_format == self.output_format and self._output_fields_inited:
            return
        if output_format:
            self.output_format = output_format
            self.output_fields.clear()
            plugin_output = plugin_manager.plugin_registry[self.output_format]
            if plugin_output.plugin_object is not None and hasattr(
                plugin_output.plugin_object, "dump"
            ):
                option_class = get_type_hints(plugin_output.plugin_object.dump)["options"]
                output_fields = self.inspect_fields(option_class)
                self.output_fields.append_many(output_fields)
            if not self._output_fields_inited:
                self._output_fields_inited = True
            self.output_fields_changed.emit()

    @Slot(str, result="QVariant")
    def plugin_info(self, name: str) -> dict[str, str]:
        assert name in {"input_format", "output_format"}
        if (suffix := getattr(self, name)) in plugin_manager.plugin_registry:
            plugin = plugin_manager.plugin_registry[suffix]
            return {
                "name": plugin.name,
                "author": plugin.author,
                "website": plugin.website,
                "description": plugin.description,
                "version": str(plugin.version),
                "file_format": plugin.file_format,
                "suffix": f"(*.{plugin.suffix})",
                "icon_base64": f"data:image/png;base64,{plugin.icon_base64}",
            }
        return {
            "name": "",
            "author": "",
            "website": "",
            "description": "",
            "version": "",
            "file_format": "",
            "suffix": "(*.*)",
            "icon_base64": "",
        }

    @Slot()
    def reset_stems(self) -> None:
        for i, task in enumerate(self.tasks):
            self.tasks.update(i, {"stem": pathlib.Path(task["path"]).stem})

    @Slot(list)
    def add_task_paths(self, paths: list[str]) -> None:
        tasks = []
        path_obj = None
        for path in paths:
            path_obj = pathlib.Path(path)
            tasks.append(
                BaseTask(
                    path=path,
                    name=path_obj.name,
                    stem=path_obj.stem,
                    ext=self.output_ext,
                    success=None,
                )
            )
        self.tasks.append_many([dataclasses.asdict(task) for task in tasks])
        if (
            settings.auto_detect_input_format
            and path_obj is not None
            and (suffix := path_obj.suffix[1:]) in plugin_manager.plugin_registry
        ):
            self.set_str("input_format", suffix)

    @Slot(str, result=QAbstractListModel)
    def qget(self, name: str) -> Any:
        return getattr(self, name)

    @Slot(str, result=QAbstractListModel)
    def get_middleware_fields(self, name: str) -> Any:
        return self.middleware_fields[name]

    @Slot(str, result=str)
    def get_str(self, name: str) -> str:
        assert name in {"input_format", "output_format"}
        return getattr(self, name)

    @Slot(str, str)
    def set_str(self, name: str, value: str) -> None:
        assert name in {"input_format", "output_format"}
        if (
            name == "input_format"
            and settings.reset_tasks_on_input_change
            and value != self.input_format
            and (
                delete_len := more_itertools.ilen(
                    more_itertools.rstrip(
                        self.tasks,
                        lambda task: task["path"].lower().endswith(f".{value}"),
                    )
                )
            )
        ):
            self.tasks.delete_many(0, delete_len)
        getattr(self, f"{name}_changed").emit(value)

    def plugin_info_file(self, plugin_archive: zipfile.Path) -> zipfile.Path | None:
        plugin_info_filename = None
        for plugin_file in plugin_archive.iterdir():
            if plugin_file.is_file() and plugin_file.name.endswith(
                f".{plugin_manager.info_extension}"
            ):
                plugin_info_filename = plugin_file
        return plugin_info_filename

    @Slot(list, result="QVariant")
    def extract_plugin_infos(self, paths: list[str]) -> list[dict[str, str]]:
        infos: list[dict[str, str]] = []
        for path in paths:
            zip_file = zipfile.Path(path)
            if (plugin_info_filename := self.plugin_info_file(zip_file)) is not None:
                plugin_info = plugin_manager.plugin_info_class.load(plugin_info_filename)
                if plugin_info is not None:
                    infos.append(
                        {
                            "name": plugin_info.name,
                            "author": plugin_info.author,
                            "version": plugin_info.version,
                            "file_format": plugin_info.file_format,
                            "suffix": f"(*.{plugin_info.suffix})",
                            "info_filename": str(plugin_info_filename),
                        }
                    )
                    logger.debug(infos)
        return infos

    @Slot()
    def check_busy(self) -> None:
        if self.busy:
            return
        self.set_busy(False)
        if self.timer.active:
            self.timer.stop()
            if settings.open_save_folder_on_completion:
                success_task = next(
                    (
                        task
                        for i, task in enumerate(self.tasks)
                        if task["success"] and not self.output_path_exists(i)
                    ),
                    None,
                )
                if success_task is not None:
                    output_dir = self.output_dir(success_task)
                    open_path(output_dir)

    def set_busy(self, busy: bool) -> None:
        self.busy_changed.emit(busy)

    @Property(bool, notify=busy_changed)
    def busy(self) -> bool:
        return self.thread_pool.active_thread_count > 0

    @Slot()
    def start(self) -> None:
        if self.input_format is None or self.output_format is None:
            error_message = "Please select input and output formats first."
            for i in range(len(self.tasks)):
                self.tasks.update(
                    i,
                    {"success": False, "error": error_message, "warning": ""},
                )
            return
        input_options = {field["name"]: field["value"] for field in self.input_fields}
        output_options = {field["name"]: field["value"] for field in self.output_fields}
        middleware_options = {
            middleware_state["identifier"]: {
                field["name"]: field["value"]
                for field in self.middleware_fields[middleware_state["identifier"]]
            }
            for middleware_state in self.middleware_states
            if middleware_state["value"]
        }
        for i in range(len(self.tasks)):
            self.tasks.update(i, {"success": False, "error": "", "warning": ""})
        self.thread_pool.max_thread_count = (
            max(len(self.tasks), 4) if settings.multi_threaded_conversion else 1
        )
        if self._conversion_mode == ConversionMode.DIRECT:
            for i in range(len(self.tasks)):
                input_path = self.tasks[i]["path"]
                output_path = f"memory:/{uuid_str()}"
                worker = ConversionWorker(
                    i,
                    input_path,
                    output_path,
                    self.input_format,
                    self.output_format,
                    input_options,
                    output_options,
                    middleware_options,
                )
                worker.signals.result.connect(
                    self.tasks.update,
                    type=Qt.ConnectionType.BlockingQueuedConnection,
                )
                self.tasks.update(
                    i,
                    {
                        "running": True,
                        "success": None,
                        "error": None,
                        "warning": None,
                    },
                )
                self.thread_pool.start(worker)
                if not i:
                    self.set_busy(True)
        elif self._conversion_mode == ConversionMode.SPLIT:
            for i in range(len(self.tasks)):
                input_path = self.tasks[i]["path"]
                output_dir = f"memory:/{uuid_str()}"
                worker = SplitWorker(
                    i,
                    input_path,
                    output_dir,
                    self.input_format,
                    self.output_format,
                    input_options,
                    output_options,
                    middleware_options,
                )
                worker.signals.result.connect(
                    self.tasks.update,
                    type=Qt.ConnectionType.BlockingQueuedConnection,
                )
                self.tasks.update(
                    i,
                    {
                        "running": True,
                        "success": None,
                        "error": None,
                        "warning": None,
                    },
                )
                self.thread_pool.start(worker)
                if not i:
                    self.set_busy(True)
        elif self._conversion_mode == ConversionMode.MERGE and len(self.tasks):
            input_paths = [task["path"] for task in self.tasks]
            output_path = f"memory:/{uuid_str()}"
            worker = MergeWorker(
                0,
                input_paths,
                output_path,
                self.input_format,
                self.output_format,
                input_options,
                output_options,
                middleware_options,
            )
            worker.signals.result.connect(
                self.tasks.update,
                type=Qt.ConnectionType.BlockingQueuedConnection,
            )
            self.tasks.update(
                0,
                {
                    "running": True,
                    "success": None,
                    "error": None,
                    "warning": None,
                },
            )
            self.thread_pool.start(worker)
            self.set_busy(True)
        self.timer.start()



import dataclasses

import pypinyin

from libresvip.model.base import (
    Note,
    Params,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point

from .model import NNInfoLine, NNNote, NNProject, NNTimeSignature
from .options import InputOptions


@dataclasses.dataclass
class NiaoNiaoParser:
    options: InputOptions
    length_multiplier: int = dataclasses.field(init=False)

    def parse_project(self, nn_project: NNProject) -> Project:
        self.length_multiplier = 60 if nn_project.info_line.version == 19 else 30
        return Project(
            song_tempo_list=self.parse_tempos(nn_project.info_line),
            time_signature_list=self.parse_time_signatures(nn_project.info_line.time_signature),
            track_list=self.parse_tracks(nn_project.notes),
        )

    def parse_tempos(self, info_line: NNInfoLine) -> list[SongTempo]:
        return [SongTempo(bpm=info_line.tempo, position=0)]

    def parse_time_signatures(self, time_signature: NNTimeSignature) -> list[TimeSignature]:
        return [
            TimeSignature(
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
        ]

    def parse_tracks(self, notes: list[NNNote]) -> list[SingingTrack]:
        return [
            SingingTrack(
                note_list=self.parse_notes(notes),
                edited_params=self.parse_params(notes),
            )
        ]

    def parse_notes(self, notes: list[NNNote]) -> list[Note]:
        note_list = []
        for nn_note in notes:
            note = Note(
                lyric=nn_note.lyric,
                start_pos=nn_note.start * self.length_multiplier,
                length=nn_note.duration * self.length_multiplier,
                key_number=88 - nn_note.key,
            )
            phonemes = pypinyin.pinyin(nn_note.lyric, heteronym=True, style=pypinyin.STYLE_NORMAL)
            if len(phonemes[0]) > 1 or phonemes[0][0] != nn_note.pronunciation:
                note.pronunciation = nn_note.pronunciation
            note_list.append(note)
        return note_list

    def parse_params(self, notes: list[NNNote]) -> Params:
        params = Params()
        for nn_note in notes:
            if (
                self.options.import_pitch
                and nn_note.pitch.point_count > 0
                and any(point != 50 for point in nn_note.pitch.points)
            ):
                step = nn_note.duration * self.length_multiplier / (nn_note.pitch.point_count - 1)
                pbs = nn_note.pitch_bend_sensitivity + 1
                params.pitch.points.append(
                    Point(nn_note.start * self.length_multiplier - 5 + 1920, -100)
                )
                for i in range(nn_note.pitch.point_count):
                    params.pitch.points.append(
                        Point(
                            round(nn_note.start * self.length_multiplier + i * step) + 1920,
                            round(
                                ((nn_note.pitch.points[i] - 50) / 50 * pbs + 88 - nn_note.key) * 100
                            ),
                        )
                    )
                params.pitch.points.append(
                    Point(
                        (nn_note.start + nn_note.duration) * self.length_multiplier + 5 + 1920,
                        -100,
                    )
                )
            # TODO: volume
        return params



from libresvip.utils.music_math import (
    cosine_easing_in_interpolation,
    cosine_easing_in_out_interpolation,
    cosine_easing_out_interpolation,
    linear_interpolation,
)


def interpolate_shape(
    start: tuple[float, float], end: tuple[float, float], x: float, shape: str
) -> float:
    if shape == "io":
        return cosine_easing_in_out_interpolation(x, start, end)
    elif shape == "i":
        return cosine_easing_in_interpolation(x, start, end)
    elif shape == "o":
        return cosine_easing_out_interpolation(x, start, end)
    else:
        return linear_interpolation(x, start, end)



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.text import to_unicode
from libresvip.utils.yamlutils import dump_yaml_1_2, load_yaml_1_2

from .model import USTXProject
from .options import InputOptions, OutputOptions
from .ustx_generator import UstxGenerator
from .ustx_parser import UstxParser


class OpenUtauConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        proj_text = to_unicode(path.read_bytes())
        ustx_project = USTXProject.model_validate(load_yaml_1_2(proj_text), context={"path": path})
        return UstxParser(options).parse_project(ustx_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ustx_project = UstxGenerator(options).generate_project(project)
        proj_dict = ustx_project.model_dump(by_alias=True, exclude_none=True)
        proj_text = dump_yaml_1_2(proj_dict)
        path.write_bytes(proj_text.encode("utf-8"))



import dataclasses
from collections.abc import Callable

from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Params,
    Phones,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
    VibratoParam,
)
from libresvip.model.point import Point
from libresvip.utils.text import CustomBoundriesBlacklist

from .models import opensvip_singers, svip_note_head_tags, svip_reverb_presets
from .msnrbf.xstudio_models import (
    XSAppModel,
    XSInstrumentTrack,
    XSITrack,
    XSLineParam,
    XSNote,
    XSNotePhoneInfo,
    XSSingingTrack,
    XSSongBeat,
    XSSongTempo,
)
from .options import InputOptions

unsupported_symbols = CustomBoundriesBlacklist(
    [",", "，", ".", "。", "?", "？", "!", "！"],
    right_boundary="$",
    match_substrings=True,
)


@dataclasses.dataclass
class BinarySvipParser:
    options: InputOptions

    def parse_project(self, version: str, model: XSAppModel) -> Project:
        project = Project()
        project.version = version
        for tempo in model.tempo_list.buf.items:
            project.song_tempo_list.append(self.parse_song_tempo(tempo))
        for beat in model.beat_list.buf.items:
            project.time_signature_list.append(self.parse_time_signature(beat))
        for track in model.track_list.items:
            if (ele := self.parse_track(track)) is not None:
                project.track_list.append(ele)
        return project

    @staticmethod
    def parse_song_tempo(tempo: XSSongTempo) -> SongTempo:
        return SongTempo(position=tempo.pos, bpm=tempo.tempo / 100.0)

    @staticmethod
    def parse_time_signature(beat: XSSongBeat) -> TimeSignature:
        frac = beat.beat_size
        return TimeSignature(
            bar_index=beat.bar_index,
            numerator=frac.x,
            denominator=frac.y,
        )

    def parse_track(self, track: XSITrack) -> Track | None:
        if isinstance(track, XSSingingTrack):
            result_track = SingingTrack()
            result_track.ai_singer_name = opensvip_singers.get_name(track.ai_singer_id)
            result_track.reverb_preset = svip_reverb_presets.inverse.get(track.reverb_preset.value)
            for note in track.note_list.buf.items:
                if (ele := self.parse_note(note)) is not None:
                    result_track.note_list.append(ele)
            result_track.edited_params = self.parse_params(track)
        elif self.options.import_instrumental_track and isinstance(track, XSInstrumentTrack):
            result_track = InstrumentalTrack(
                audio_file_path=track.instrument_file_path,
                offset=track.offset_in_pos,
            )
        else:
            return None
        result_track.title = track.name
        result_track.mute = track.mute
        result_track.solo = track.solo
        result_track.volume = track.volume
        result_track.pan = track.pan
        return result_track

    def parse_params(self, track: XSSingingTrack) -> Params:
        params = Params()
        if self.options.import_pitch and (pitch_line := track.edited_pitch_line) is not None:
            params.pitch = self.parse_param_curve(
                pitch_line, op=lambda x: x - 1150 if x > 1050 else -100
            )
        if self.options.import_volume and (volume_line := track.edited_volume_line) is not None:
            params.breath = self.parse_param_curve(volume_line)
        if self.options.import_breath and (breath_line := track.edited_breath_line) is not None:
            params.breath = self.parse_param_curve(breath_line)
        if self.options.import_gender and (gender_line := track.edited_gender_line) is not None:
            params.gender = self.parse_param_curve(gender_line)
        if self.options.import_strength and (power_line := track.edited_power_line) is not None:
            params.strength = self.parse_param_curve(power_line)
        return params

    def parse_note(self, note: XSNote) -> Note:
        result_note = Note(
            start_pos=note.start_pos,
            length=note.width_pos,
            key_number=note.key_index - 12,
            head_tag=svip_note_head_tags.inverse.get(note.head_tag.value),
            lyric=unsupported_symbols.cleanse_text(note.lyric),
        )
        if pronunciation := note.pronouncing:
            result_note.pronunciation = pronunciation
        if (phone := note.note_phone_info) is not None:
            result_note.edited_phones = self.parse_phones(phone)
        if note.vibrato is not None:
            result_note.vibrato = self.parse_vibrato(note)
        return result_note

    def parse_vibrato(self, note: XSNote) -> VibratoParam:
        kwargs = {}
        if note.vibrato_percent_info is not None:
            kwargs["start_percent"] = note.vibrato_percent_info.start_percent
            kwargs["end_percent"] = note.vibrato_percent_info.end_percent
        elif note.vibrato_percent > 0:
            kwargs["start_percent"] = 1.0 - note.vibrato_percent / 100.0
            kwargs["end_percent"] = 1.0
        if note.vibrato is not None:
            kwargs["is_anti_phase"] = note.vibrato.is_anti_phase
            kwargs["amplitude"] = self.parse_param_curve(note.vibrato.amp_line)
            kwargs["frequency"] = self.parse_param_curve(note.vibrato.freq_line)
        return VibratoParam(**kwargs)

    @staticmethod
    def parse_phones(phone: XSNotePhoneInfo) -> Phones:
        return Phones(
            head_length_in_secs=phone.head_phone_time_in_sec,
            mid_ratio_over_tail=phone.mid_part_over_tail_part_ratio,
        )

    @staticmethod
    def parse_param_curve(
        line: XSLineParam, op: Callable[[float], float] | None = None
    ) -> ParamCurve:
        if op is None:

            def _op(x: float) -> float:
                return x

            op = _op

        param_curve = ParamCurve()
        for point in line.nodes:
            param_curve.points.append(Point(x=point.pos, y=int(op(point.value))))
        return param_curve



import pathlib
from typing import TYPE_CHECKING, Any

from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata_pydantic.bindings import XmlParser, XmlSerializer

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project
from libresvip.utils.xmlutils import DefaultXmlWriter

from .models.enums import VsqxVersion
from .models.vsqx3 import VSQ3_NS
from .models.vsqx4 import VSQ4_NS
from .options import InputOptions, OutputOptions
from .vsq3_generator import Vsq3Generator
from .vsq4_generator import Vsq4Generator
from .vsqx_parser import VsqxParser

if TYPE_CHECKING:
    from .model import Vsqx


class VocaloidXMLWriter(DefaultXmlWriter):
    def set_data(self, data: Any) -> None:
        if (
            isinstance(data, str)
            and self.pending_tag
            and len(self.pending_tag) > 1
            and self.pending_tag[1]
            in (
                "y",
                "p",
                "id",
                "id2",
                "auxID",
                "compID",
                "stylePluginID",
                "vstPluginID",
                "name",
                "partName",
                "seqName",
                "stylePluginName",
                "trackName",
                "vVoiceName",
                "vstPluginName",
                "filePath",
                "content",
                "phnmStr",
                "vender",
                "comment",
                "version",
            )
        ):
            super().set_cdata(data)
        else:
            super().set_data(data)

    def start_document(self) -> None:
        if self.config.xml_declaration:
            self.output.write(
                f'<?xml version="{self.config.xml_version}" encoding="{self.config.encoding}" standalone="no"?>\n'
            )


class VsqxConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        xml_parser = XmlParser()
        vsqx_proj: Vsqx = xml_parser.from_bytes(path.read_bytes())
        return VsqxParser(options, path).parse_project(vsqx_proj)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        vsqx_generator_class: type[Vsq3Generator | Vsq4Generator]
        if options.vsqx_version == VsqxVersion.VSQ3:
            vsqx_generator_class = Vsq3Generator
            vsqx_namespace = VSQ3_NS
        else:
            vsqx_generator_class = Vsq4Generator
            vsqx_namespace = VSQ4_NS
        vsqx_proj = vsqx_generator_class(options).generate_project(project)
        xml_serializer = XmlSerializer(
            config=SerializerConfig(
                pretty_print=options.pretty_xml,
                pretty_print_indent="\t",
                schema_location=f"{vsqx_namespace} vsq{options.vsqx_version.value}.xsd",
            ),
            writer=VocaloidXMLWriter,
        )
        path.write_bytes(
            xml_serializer.render(vsqx_proj, ns_map={None: vsqx_namespace}).encode("utf-8")
        )



from enum import Enum
from typing import Final

VELOCITY_CONSTANT: Final[float] = 1.75
EXPRESSION_CONSTANT: Final[float] = 2.0


class ControlChange(Enum):
    BANK_SELECT = 0
    MODULATION = 1
    BREATH = 2
    FOOT = 4
    PORTAMENTO_TIME = 5
    DATA_ENTRY = 6
    VOLUME = 7
    BALANCE = 8
    PAN = 10
    EXPRESSION = 11
    EFFECT_1 = 12
    EFFECT_2 = 13
    GENERAL_1 = 16
    GENERAL_2 = 17
    GENERAL_3 = 18
    GENERAL_4 = 19
    BANK_SELECT_L = 32
    MODULATION_L = 33
    BREATH_L = 34
    FOOT_L = 36
    PORTAMENTO_TIME_L = 37
    DATA_ENTRY_L = 38
    VOLUME_L = 39
    BALANCE_L = 40
    PAN_L = 42
    EXPRESSION_L = 43
    EFFECT_1_L = 44
    EFFECT_2_L = 45
    GENERAL_1_L = 48
    GENERAL_2_L = 49
    GENERAL_3_L = 50
    GENERAL_4_L = 51
    HOLD = 64
    PORTAMENTO = 65
    SUSTENUTO = 66
    SOFT_PEDAL = 67
    LEGATO_FOOTSWITCH = 68
    HOLD_2 = 69
    SOUND_VARIATION = 70
    SOUND_TIMBRE = 71
    SOUND_RELEASE_TIME = 72
    SOUND_ATTACK_TIME = 73
    SOUND_BRIGHTNESS = 74
    SOUND_CONTROL_6 = 75
    SOUND_CONTROL_7 = 76
    SOUND_CONTROL_8 = 77
    SOUND_CONTROL_9 = 78
    SOUND_CONTROL_10 = 79
    GENERAL_5 = 80
    GENERAL_6 = 81
    GENERAL_7 = 82
    GENERAL_8 = 83
    PORTAMENTO_CONTROL = 84
    EFFECTS_DEPTH_1 = 91
    EFFECTS_DEPTH_2 = 92
    EFFECTS_DEPTH_3 = 93
    EFFECTS_DEPTH_4 = 94
    EFFECTS_DEPTH_5 = 95
    DATA_INCREMENT = 96
    DATA_DECREMENT = 97
    NRPN_LSB = 98
    NRPN_MSB = 99
    RPN_LSB = 100
    RPN_MSB = 101
    ALL_SOUND_OFF = 120
    RESET_ALL_CONTROLLERS = 121
    LOCAL_CONTROL = 122
    ALL_NOTES_OFF = 123
    OMNI_MODE_OFF = 124
    OMNI_MODE_ON = 125
    MONO_MODE_ON = 126
    POLY_MODE_ON = 127



import sys

from libresvip.core.warning_types import warning_logger
from libresvip.extension.manager import get_translation
from libresvip.utils import translation

warning_logger.add(sys.stderr, format="{message}", level="WARNING")
translation.singleton_translation = get_translation()



from pydantic import Field

from libresvip.core.constants import DEFAULT_BPM
from libresvip.model.base import BaseModel


class Y77Note(BaseModel):
    py: str = ""
    length: int = Field(alias="len")
    start: int
    lyric: str
    pitch: int
    pbs: int = 0
    pit: list[int | float] = Field(default_factory=list)


class Y77Project(BaseModel):
    bpm: float = DEFAULT_BPM
    bars: int = 0
    notes: list[Y77Note] = Field(default_factory=list)
    nnote: int = 0
    bbar: int = 4
    v: int = 10001
    bbeat: int = 4



import dataclasses

import more_itertools

from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.point import Point
from libresvip.utils.music_math import hz2midi

from .model import (
    VoiceVoxNote,
    VoiceVoxProject,
    VoiceVoxTempo,
    VoiceVoxTimeSignature,
    VoiceVoxTrack,
)
from .options import InputOptions


@dataclasses.dataclass
class VOICEVOXParser:
    options: InputOptions
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def parse_project(self, voicevox_project: VoiceVoxProject) -> Project:
        time_signatures = self.parse_time_signatures(voicevox_project.song.time_signatures)
        self.first_bar_length = int(time_signatures[0].bar_length(voicevox_project.song.tpqn))
        tempos = self.parse_tempos(voicevox_project.song.tempos)
        self.time_synchronizer = TimeSynchronizer(tempos)
        return Project(
            time_signature_list=time_signatures,
            song_tempo_list=tempos,
            track_list=[
                self.parse_track(voicevox_project.song.tracks[track_key])
                for track_key in voicevox_project.song.track_order
            ],
        )

    def parse_time_signatures(
        self, time_signatures: list[VoiceVoxTimeSignature]
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=time_signature.measure_number - 1,
                numerator=time_signature.beats,
                denominator=time_signature.beat_type,
            )
            for time_signature in time_signatures
        ]

    def parse_tempos(self, tempos: list[VoiceVoxTempo]) -> list[SongTempo]:
        return [SongTempo(bpm=tempo.bpm, position=tempo.position) for tempo in tempos]

    def parse_track(self, track: VoiceVoxTrack) -> SingingTrack:
        singing_track = SingingTrack(
            title=track.name,
            solo=track.solo,
            mute=track.mute,
            volume=track.gain,
            pan=track.pan,
            note_list=[self.parse_note(note) for note in track.notes],
        )
        if self.options.import_pitch:
            singing_track.edited_params.pitch.points.root = self.parse_pitch(track.pitch_edit_data)
        return singing_track

    def parse_note(self, note: VoiceVoxNote) -> Note:
        return Note(
            start_pos=note.position,
            length=note.duration,
            lyric=note.lyric,
            key_number=note.note_number,
        )

    def parse_pitch(self, pitch_edit_data: list[float]) -> list[Point]:
        secs = 0.0
        secs_step = 4 / 375
        points = [Point.start_point()]
        for part in more_itertools.split_when(
            pitch_edit_data,
            lambda x, y: [x, y].count(-1) == 1,
        ):
            if all(value == -1 for value in part):
                secs += secs_step * len(part)
            else:
                points.append(
                    Point(
                        int(self.time_synchronizer.get_actual_ticks_from_secs(secs))
                        + self.first_bar_length,
                        -100,
                    )
                )
                for value in part:
                    points.append(
                        Point(
                            int(self.time_synchronizer.get_actual_ticks_from_secs(secs))
                            + self.first_bar_length,
                            round(hz2midi(value) * 100),
                        )
                    )
                    secs += secs_step
                points.append(
                    Point(
                        int(self.time_synchronizer.get_actual_ticks_from_secs(secs))
                        + self.first_bar_length,
                        -100,
                    )
                )
        points.append(Point.end_point())
        return points



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class ProjectZoomOptions(Enum):
    NONE: Annotated[str, Field(title=_("None"))] = "1/1"
    DOUBLE: Annotated[str, Field(title="2")] = "2/1"
    HALF: Annotated[str, Field(title=_("1/2"))] = "1/2"
    FIVE_THIRDS: Annotated[str, Field(title=_("5/3"))] = "5/3"
    THREE_HALVES: Annotated[str, Field(title=_("3/2"))] = "3/2"
    SIX_FIFTHS: Annotated[str, Field(title=_("6/5"))] = "6/5"
    FOUR_FIFTHS: Annotated[str, Field(title=_("4/5"))] = "4/5"
    THREE_FIFTHS: Annotated[str, Field(title=_("3/5"))] = "3/5"
    THREE_QUARTER: Annotated[str, Field(title=_("3/4"))] = "3/4"


class ProcessOptions(BaseModel):
    factor: ProjectZoomOptions = Field(
        default=ProjectZoomOptions.NONE,
        title=_("Zoom factor"),
        description=_(
            "Change Bpm and notes in parallel so that the actual singing speed is kept. For example, with factor 2, 60 bpm becomes 120 bpm and all notes become twice the length"
        ),
    )



import configparser
import dataclasses
import math
import re

import mido_fix as mido

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)

from .options import BreathOption, InputOptions
from .vocaloid_pitch import (
    ControllerEvent,
    VocaloidPartPitchData,
    pitch_from_vocaloid_parts,
)

BREATH_PATTERN = re.compile("br[1-5]")


@dataclasses.dataclass
class VsqParser:
    options: InputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    ticks_per_beat: int = dataclasses.field(init=False)

    @property
    def tick_rate(self) -> float:
        return TICKS_IN_BEAT / self.ticks_per_beat

    def parse_project(self, vsq_project: mido.MidiFile) -> Project:
        self.ticks_per_beat = vsq_project.ticks_per_beat
        self._convert_delta_to_cumulative(vsq_project.tracks)
        tracks_as_text = self.extract_vsq_text_from_meta_events(vsq_project.tracks)
        measure_prefix = self.get_measure_prefix(tracks_as_text[0])
        master_track = vsq_project.tracks[0]
        time_signature_list = self.parse_time_signatures(master_track)
        self.first_bar_length = round(time_signature_list[0].bar_length(self.ticks_per_beat))
        tick_prefix = self.first_bar_length * measure_prefix
        song_tempo_list = self.parse_tempo(master_track, tick_prefix)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=time_signature_list,
            track_list=[
                self.parse_track(text, i + 1, tick_prefix) for i, text in enumerate(tracks_as_text)
            ],
        )

    @staticmethod
    def get_measure_prefix(text: str) -> int:
        master_parser = configparser.ConfigParser()
        try:
            master_parser.read_string(text)
        except configparser.Error:
            master_parser.read_string(text.rsplit("\n", 1)[0])
        return master_parser.getint("Master", "PreMeasure", fallback=1)

    @staticmethod
    def extract_vsq_text_from_meta_events(
        tracks: list[mido.MidiTrack],
    ) -> list[str]:
        return [
            text
            for track in tracks
            if (
                text := "".join(
                    event.text.removeprefix("DM:").partition(":")[-1]
                    for event in track
                    if isinstance(event, mido.MetaMessage) and event.type == "text"
                )
            )
        ]

    @staticmethod
    def _convert_delta_to_cumulative(tracks: list[mido.MidiTrack]) -> None:
        for track in tracks:
            tick = 0
            for event in track:
                event.time += tick
                tick = event.time

    def parse_time_signatures(self, master_track: mido.MidiTrack) -> list[TimeSignature]:
        time_signature_changes: list[TimeSignature] = []

        # traversing
        prev_ticks = 0
        measure = 0
        for event in master_track:
            if event.type == "time_signature":
                tick_in_full_note = (
                    time_signature_changes[-1].bar_length(self.ticks_per_beat)
                    if time_signature_changes
                    else 4 * self.ticks_per_beat
                )
                tick = event.time
                measure += (tick - prev_ticks) / tick_in_full_note
                ts_obj = TimeSignature(
                    bar_index=max(math.floor(measure), 0),
                    numerator=event.numerator,
                    denominator=event.denominator,
                )
                time_signature_changes.append(ts_obj)
                prev_ticks = tick
        if not time_signature_changes:
            time_signature_changes.append(TimeSignature(bar_index=0, numerator=4, denominator=4))
        return time_signature_changes

    def parse_tempo(self, master_track: mido.MidiTrack, tick_prefix: int) -> list[SongTempo]:
        tempos = []

        # traversing
        for event in master_track:
            if event.type == "set_tempo":
                # convert tempo to BPM
                tempo = round(mido.tempo2bpm(event.tempo), 3)
                tick = round(event.time * self.tick_rate)
                if tick == 0:
                    tempos = [SongTempo(position=0, bpm=tempo)]
                else:
                    last_tempo = tempos[-1].bpm
                    if tempo != last_tempo:
                        tempos.append(SongTempo(position=tick - tick_prefix, bpm=tempo))
        if not tempos:
            tempos.append(SongTempo(position=0))
        return tempos

    def parse_track(self, text: str, track_index: int, tick_prefix: int) -> SingingTrack:
        vsq_track = configparser.ConfigParser()
        try:
            vsq_track.read_string(text)
        except configparser.Error:
            vsq_track.read_string(text.rsplit("\n", 1)[0])
        singer_name = ""
        if (
            singer_icon_key := next(
                (
                    value.get("iconhandle")
                    for key, value in vsq_track.items()
                    if key.startswith("ID#") and value.get("type") == "Singer"
                ),
                None,
            )
        ) is not None:
            singer_name = vsq_track.get(singer_icon_key, "ids", fallback="")
        singing_track = SingingTrack(
            ai_singer_name=singer_name,
            title=vsq_track.get("Common", "Name", fallback=f"Track {track_index}"),
            note_list=self.parse_notes(vsq_track, tick_prefix)
            if vsq_track.has_section("EventList")
            else [],
        )
        if self.options.import_pitch and (
            pitch := self.parse_pitch(vsq_track, singing_track.note_list, tick_prefix)
        ):
            singing_track.edited_params.pitch = pitch
        return singing_track

    def parse_pitch(
        self,
        vsq_track: configparser.ConfigParser,
        note_list: list[Note],
        tick_prefix: int,
    ) -> ParamCurve | None:
        pit: list[ControllerEvent] = []
        pbs: list[ControllerEvent] = []
        if vsq_track.has_section("PitchBendBPList"):
            pit.extend(
                ControllerEvent(
                    pos=int(key) - tick_prefix,
                    value=int(value),
                )
                for key, value in vsq_track["PitchBendBPList"].items()
            )
        if vsq_track.has_section("PitchBendSensBPList"):
            pbs.extend(
                ControllerEvent(
                    pos=int(key) - tick_prefix,
                    value=int(value),
                )
                for key, value in vsq_track["PitchBendSensBPList"].items()
            )
        return pitch_from_vocaloid_parts(
            [
                VocaloidPartPitchData(
                    start_pos=0,
                    pit=pit,
                    pbs=pbs,
                )
            ],
            self.synchronizer,
            note_list,
            self.first_bar_length,
        )

    def parse_notes(self, vsq_track: configparser.ConfigParser, tick_prefix: int) -> list[Note]:
        notes = []
        for tick_str, event_key in vsq_track["EventList"].items():
            tick = int(tick_str) - tick_prefix
            if event_key.startswith("ID#"):
                vsq_note = vsq_track[event_key]
                if (
                    vsq_note["type"] == "Anote"
                    and (length := vsq_note.getint("length"))
                    and (key := vsq_note.getint("note#"))
                ):
                    lyric_handle = vsq_note.get("lyrichandle") or ""
                    lyric_value, phoneme_value = vsq_track.get(
                        lyric_handle, "L0", fallback=","
                    ).split(",")[:2]
                    if BREATH_PATTERN.fullmatch(phoneme_value.strip('"')) is not None:
                        if self.options.breath == BreathOption.IGNORE:
                            continue
                        else:
                            lyric_value = phoneme_value
                    notes.append(
                        Note(
                            start_pos=tick,
                            length=length,
                            key_number=key,
                            lyric=lyric_value.strip('"'),
                            pronunciation=None,
                        )
                    )
        return notes



import dataclasses

from construct import (
    Byte,
    Bytes,
    BytesInteger,
    Const,
    Construct,
    FocusedSeq,
    GreedyBytes,
    If,
    Int16ul,
    PaddedString,
    PascalString,
    Prefixed,
    PrefixedArray,
    this,
)
from construct_typed import (
    DataclassMixin,
    DataclassStruct,
    EnumBase,
    TEnum,
    csfield,
)

Int32ul = BytesInteger(4, swapped=True)
Int32sl = BytesInteger(4, swapped=True, signed=True)
LineBreak = Const(b"\n")


def muta_prefixed_array(data_model: DataclassMixin) -> Construct:
    return PrefixedArray(
        FocusedSeq("count", LineBreak, "count" / Int32ul),
        FocusedSeq("item", LineBreak, "item" / DataclassStruct(data_model)),
    )


class MutaTrackType(EnumBase):
    TALK = 1
    SONG = 2
    AUDIO = 3


@dataclasses.dataclass
class MutaTempo(DataclassMixin):
    position: int = csfield(Int32sl)
    bpm: int = csfield(Int32ul)


@dataclasses.dataclass
class MutaTimeSignature(DataclassMixin):
    measure_position: int = csfield(Int32sl)
    numerator: int = csfield(Int32ul)
    denominator: int = csfield(Int32ul)


@dataclasses.dataclass
class MutaAudioTrackData(DataclassMixin):
    line_break: bytes = csfield(LineBreak)
    start: int = csfield(Int32ul)
    length: int = csfield(Int32ul)
    file_path: str = csfield(PascalString(Int32ul, "utf-8"))


@dataclasses.dataclass
class MutaNoteTiming(DataclassMixin):
    ori_pos: int = csfield(Int32sl)
    mod_pos: int = csfield(Int32sl)


@dataclasses.dataclass
class MutaNote(DataclassMixin):
    start: int = csfield(Int32sl)
    length: int = csfield(Int32ul)
    key: int = csfield(Int32ul)
    lyric: list[int] = csfield(Int16ul[8])
    phoneme: str = csfield(PaddedString(16, "utf-16-le"))
    tmg_data: list[MutaNoteTiming] = csfield(DataclassStruct(MutaNoteTiming)[5])


@dataclasses.dataclass
class MutaPoint(DataclassMixin):
    time: int = csfield(Int32sl)
    value: int = csfield(Int32sl)


MutaPoints = muta_prefixed_array(MutaPoint)


@dataclasses.dataclass
class MutaParams(DataclassMixin):
    unknown_param: list[MutaPoint] = csfield(MutaPoints)
    pitch_range: list[MutaPoint] = csfield(MutaPoints)
    volume_data: list[MutaPoint] = csfield(MutaPoints)
    pitch_data: list[MutaPoint] = csfield(MutaPoints)
    vibrato_amplitude_range: list[MutaPoint] = csfield(MutaPoints)
    vibrato_amplitude_data: list[MutaPoint] = csfield(MutaPoints)
    vibrato_frequency_range: list[MutaPoint] = csfield(MutaPoints)
    vibrato_frequency_data: list[MutaPoint] = csfield(MutaPoints)


@dataclasses.dataclass
class MutaSongTrackData(DataclassMixin):
    line_break: bytes = csfield(LineBreak)
    start: int = csfield(Int32ul)
    length: int = csfield(Int32ul)
    singer_name: list[int] = csfield(Int16ul[258])
    unknown_1: int = csfield(Int32ul)
    notes: list[MutaNote] = csfield(muta_prefixed_array(MutaNote))
    params: MutaParams = csfield(DataclassStruct(MutaParams))


@dataclasses.dataclass
class MutaPhoneme(DataclassMixin):
    volume: int = csfield(Byte)
    pos1: int = csfield(Int16ul)
    length: int = csfield(Byte)
    pos2: int = csfield(Int16ul)
    pitch: int = csfield(Int32ul)
    phoneme: bytes = csfield(Prefixed(Int32ul, GreedyBytes))


@dataclasses.dataclass
class MutaText(DataclassMixin):
    text: bytes = csfield(Prefixed(Int32ul, GreedyBytes))
    line_break: bytes = csfield(LineBreak)
    volume: int = csfield(Byte)
    speed: int = csfield(Byte)
    tone: int = csfield(Byte)
    quality: int = csfield(Byte)
    normal: int = csfield(Byte)
    happiness: int = csfield(Byte)
    sadness: int = csfield(Byte)
    reserved: int = csfield(Byte)
    phonemes: list[MutaPhoneme] = csfield(muta_prefixed_array(MutaPhoneme))


@dataclasses.dataclass
class MutaTalkTrackData(DataclassMixin):
    line_break1: bytes = csfield(LineBreak)
    start: int = csfield(Int32ul)
    length: int = csfield(Int32ul)
    talker_name: list[int] = csfield(Int16ul[258])
    unknown_1: int = csfield(Int32ul)
    line_break2: bytes = csfield(LineBreak)
    text: MutaText = csfield(DataclassStruct(MutaText))


@dataclasses.dataclass
class MutaTrack(DataclassMixin):
    track_type: MutaTrackType = csfield(TEnum(Int32ul, MutaTrackType))
    seq_count: int = csfield(Int32ul)
    line_break: bytes = csfield(LineBreak)
    mute: int = csfield(Int32ul)
    solo: int = csfield(Int32ul)
    volume: int = csfield(Int32ul)
    pan: int = csfield(Int32ul)
    name: str = csfield(PaddedString(12, "utf-16-le"))
    padding: bytes = csfield(Bytes(52))
    talk_track_data: list[MutaTalkTrackData] | None = csfield(
        If(
            this.track_type == MutaTrackType.TALK,
            DataclassStruct(MutaTalkTrackData)[this.seq_count],
        )
    )
    audio_track_data: list[MutaAudioTrackData] | None = csfield(
        If(
            this.track_type == MutaTrackType.AUDIO,
            DataclassStruct(MutaAudioTrackData)[this.seq_count],
        ),
    )
    song_track_data: list[MutaSongTrackData] | None = csfield(
        If(
            this.track_type == MutaTrackType.SONG,
            DataclassStruct(MutaSongTrackData)[this.seq_count],
        )
    )


@dataclasses.dataclass
class MutaProject(DataclassMixin):
    file_version: int = csfield(Int32ul)
    tempos: list[MutaTempo] = csfield(muta_prefixed_array(MutaTempo))
    time_signatures: list[MutaTimeSignature] = csfield(muta_prefixed_array(MutaTimeSignature))
    tracks: list[MutaTrack] = csfield(muta_prefixed_array(MutaTrack))
    line_break: bytes = csfield(LineBreak)


muta_project_struct = DataclassStruct(MutaProject)



from __future__ import annotations

import contextvars
import enum
import locale
import pathlib
import re
from typing import Annotated, Any, TypeVar

import pydantic_settings
import yaml
from pydantic import BaseModel, Field, GetCoreSchemaHandler, ValidationError
from pydantic_core import core_schema

try:
    from yaml import CSafeLoader as DefaultSafeLoader
except ImportError:
    from yaml import SafeLoader as DefaultSafeLoader

from typing_extensions import Self

from libresvip.core.constants import app_dir

E = TypeVar("E", bound=enum.Enum)
YAML_BOOL_TYPES = [
    "y",
    "Y",
    "yes",
    "Yes",
    "YES",
    "n",
    "N",
    "no",
    "No",
    "NO",
    "true",
    "True",
    "TRUE",
    "false",
    "False",
    "FALSE",
    "on",
    "On",
    "ON",
    "off",
    "Off",
    "OFF",
]


def pydantic_enum(enum_cls: type[E]) -> type[E]:
    def _get_pydantic_core_schema_(
        cls: type[E], source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.WrapValidatorFunctionSchema:
        assert source_type is cls

        def get_enum(value: Any, validate_next: core_schema.ValidatorFunctionWrapHandler) -> E:
            if isinstance(value, cls):
                return value
            else:
                name: str = validate_next(value)
                return enum_cls[name]

        def serialize(enum: E) -> str:
            return enum.name

        expected = [member.name for member in cls]
        name_schema = core_schema.literal_schema(expected)

        return core_schema.no_info_wrap_validator_function(
            get_enum,
            name_schema,
            ref=cls.__name__,
            serialization=core_schema.plain_serializer_function_ser_schema(serialize),
        )

    setattr(enum_cls, "__get_pydantic_core_schema__", classmethod(_get_pydantic_core_schema_))
    return enum_cls


def get_omega_conf_loader() -> DefaultSafeLoader:
    class OmegaConfLoader(DefaultSafeLoader):
        def construct_mapping(self, node: yaml.Node, deep: bool = False) -> Any:
            keys = set()
            for key_node, _ in node.value:
                if key_node.tag != yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG:
                    continue
                if key_node.value in keys:
                    msg = "while constructing a mapping"
                    raise yaml.constructor.ConstructorError(
                        msg,
                        node.start_mark,
                        f"found duplicate key {key_node.value}",
                        key_node.start_mark,
                    )
                keys.add(key_node.value)
            return super().construct_mapping(node, deep=deep)

    loader_class = OmegaConfLoader
    loader_class.add_implicit_resolver(
        "tag:yaml.org,2002:float",
        re.compile(
            """^(?:
         [-+]?[0-9]+(?:_[0-9]+)*\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?[0-9]+(?:_[0-9]+)*(?:[eE][-+]?[0-9]+)
        |\\.[0-9]+(?:_[0-9]+)*(?:[eE][-+][0-9]+)?
        |[-+]?[0-9]+(?:_[0-9]+)*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$""",
            re.X,
        ),
        list("-+0123456789."),
    )
    loader_class.yaml_implicit_resolvers = {
        key: [(tag, regexp) for tag, regexp in resolvers if tag != "tag:yaml.org,2002:timestamp"]
        for key, resolvers in loader_class.yaml_implicit_resolvers.items()
    }

    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib.Path",
        lambda loader, node: pathlib.Path(*loader.construct_sequence(node)),
    )
    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib.PosixPath",
        lambda loader, node: pathlib.PosixPath(*loader.construct_sequence(node)),
    )
    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib.WindowsPath",
        lambda loader, node: pathlib.WindowsPath(*loader.construct_sequence(node)),
    )
    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib._local.Path",
        lambda loader, node: pathlib.Path(*loader.construct_sequence(node)),
    )
    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib._local.PosixPath",
        lambda loader, node: pathlib.PosixPath(*loader.construct_sequence(node)),
    )
    loader_class.add_constructor(
        "tag:yaml.org,2002:python/object/apply:pathlib._local.WindowsPath",
        lambda loader, node: pathlib.WindowsPath(*loader.construct_sequence(node)),
    )
    return loader_class


def yaml_is_bool(st: str) -> bool:
    return st in YAML_BOOL_TYPES


def is_float(st: str) -> bool:
    try:
        float(st)
    except ValueError:
        return False
    else:
        return True


def is_int(st: str) -> bool:
    try:
        int(st)
    except ValueError:
        return False
    else:
        return True


class OmegaConfDumper(yaml.Dumper):
    str_representer_added = False

    @staticmethod
    def str_representer(dumper: yaml.Dumper, data: str) -> yaml.ScalarNode:
        with_quotes = yaml_is_bool(data) or is_int(data) or is_float(data)
        return dumper.represent_scalar(
            yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
            data,
            style=("'" if with_quotes else None),
        )


def get_omega_conf_dumper() -> type[OmegaConfDumper]:
    if not OmegaConfDumper.str_representer_added:
        OmegaConfDumper.add_representer(str, OmegaConfDumper.str_representer)
        OmegaConfDumper.str_representer_added = True
    return OmegaConfDumper


class YamlSettings(pydantic_settings.BaseSettings):
    model_config = pydantic_settings.SettingsConfigDict(extra="allow", validate_assignment=True)
    _settings_dir: pathlib.Path
    __FILENAME__: str = "settings.yml"

    @classmethod
    def exists(cls, settings_dir: str | pathlib.Path) -> bool:
        return (pathlib.Path(settings_dir).resolve() / cls.__FILENAME__).exists()

    @classmethod
    def create(cls, settings_dir: str | pathlib.Path, exists_ok: bool = False) -> Self:
        settings_dir = pathlib.Path(settings_dir).resolve()
        if not exists_ok and cls.exists(settings_dir):
            msg = f"`{cls.__FILENAME__}` already exists in `{settings_dir}`"
            raise FileExistsError(msg)
        settings = cls()
        settings._settings_dir = settings_dir
        settings.save()
        return settings

    @classmethod
    def load(
        cls,
        settings_dir: str | pathlib.Path,
        create_if_missing: bool = False,
        raise_error_if_failed: bool = True,
    ) -> Self:
        settings_dir = pathlib.Path(settings_dir).resolve()
        if not cls.exists(settings_dir):
            if create_if_missing:
                return cls.create(settings_dir)
            msg = f"`{cls.__FILENAME__}` not found in `{settings_dir}`"
            raise FileNotFoundError(msg)

        settings_data = yaml.load(
            (settings_dir / cls.__FILENAME__).read_text(encoding="utf-8"), get_omega_conf_loader()
        )
        try:
            settings_object = cls.model_validate(settings_data)
        except ValidationError as e:
            msg = f"Invalid settings data: {e}"
            if raise_error_if_failed:
                raise ValueError(msg)
            settings_object = cls()
        settings_object._settings_dir = settings_dir
        return settings_object

    def save(self) -> None:
        file_path = self._settings_dir / self.__FILENAME__
        try:
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(
                yaml.dump(self.model_dump(), None, get_omega_conf_dumper()), encoding="utf-8"
            )
        except OSError as e:
            msg = f"Failed to save settings to {file_path}: {e}"
            raise OSError(msg)


class LyricsReplaceMode(enum.Enum):
    FULL = "full"
    ALPHABETIC = "alphabetic"
    NON_ALPHABETIC = "non_alphabetic"
    REGEX = "regex"


class LyricsReplacement(BaseModel):
    replacement: str
    pattern_main: str
    pattern_prefix: str = ""
    pattern_suffix: str = ""
    flags: Annotated[re.RegexFlag, pydantic_enum(re.RegexFlag)] = re.IGNORECASE
    mode: Annotated[LyricsReplaceMode, pydantic_enum(LyricsReplaceMode)] = LyricsReplaceMode.FULL

    def __post_init__(self) -> None:
        if self.mode == LyricsReplaceMode.FULL:
            self.pattern_prefix = "^"
            self.pattern_suffix = "$"
        elif self.mode == LyricsReplaceMode.ALPHABETIC:
            self.pattern_prefix = r"(?<=^|\b)"
            self.pattern_suffix = r"(?=$|\b)"
        elif self.mode == LyricsReplaceMode.NON_ALPHABETIC:
            self.pattern_prefix = self.pattern_suffix = ""
        if self.mode != LyricsReplaceMode.REGEX:
            self.pattern_main = re.escape(self.pattern_main)
        else:
            try:
                self.compiled_pattern
            except re.error as e:
                msg = f"Invalid pattern: {self._pattern}"
                raise ValueError(msg) from e

    @property
    def compiled_pattern(self) -> re.Pattern[str]:
        return re.compile(self._pattern, self.flags)

    @property
    def _pattern(self) -> str:
        return f"{self.pattern_prefix}{self.pattern_main}{self.pattern_suffix}"

    def replace(self, text: str) -> str:
        return self.compiled_pattern.sub(self.replacement, text)


class Language(enum.Enum):
    CHINESE = "zh_CN"
    ENGLISH = "en_US"
    JAPANESE = "ja_JP"
    GERMAN = "de_DE"

    @staticmethod
    def to_language(locale: str) -> str:
        """Turn a locale name (en_US) into a language name (en-us)."""
        p = locale.find("_")
        if p >= 0:
            return f"{locale[:p].lower()}-{locale[p + 1 :].lower()}"
        else:
            return locale.lower()

    @classmethod
    def from_locale(cls, locale: str) -> Language:
        locale = cls.to_language(locale)
        return {
            "zh-cn": cls.CHINESE,
            "ja-jp": cls.JAPANESE,
            "de-de": cls.GERMAN,
        }.get(locale, cls.ENGLISH)

    @classmethod
    def auto(cls) -> Language:
        sys_locale = locale.getdefaultlocale()[0]
        return cls.from_locale(sys_locale or "en_US")


class DarkMode(enum.Enum):
    LIGHT = "Light"
    DARK = "Dark"
    SYSTEM = "System"


class ConflictPolicy(enum.Enum):
    SKIP = "Skip"
    OVERWRITE = "Overwrite"
    PROMPT = "Prompt"


class ConversionMode(enum.Enum):
    DIRECT = "Direct"
    SPLIT = "Split"
    MERGE = "Merge"


class LibreSvipBaseUISettings(YamlSettings):
    language: Annotated[Language, pydantic_enum(Language)] = Field(default_factory=Language.auto)
    last_input_format: str | None = Field(default=None)
    last_output_format: str | None = Field(default=None)
    dark_mode: Annotated[DarkMode, pydantic_enum(DarkMode)] = Field(default=DarkMode.SYSTEM)
    auto_detect_input_format: bool = Field(default=True)
    reset_tasks_on_input_change: bool = Field(default=True)
    max_track_count: int = Field(default=1)
    lyric_replace_rules: dict[str, list[LyricsReplacement]] = Field(default_factory=dict)


ui_settings_ctx: contextvars.ContextVar[LibreSvipBaseUISettings | None] = contextvars.ContextVar(
    "ui_settings_ctx"
)


class LibreSvipSettings(LibreSvipBaseUISettings):
    disabled_plugins: list[str] = Field(default_factory=list)
    # GUI Only
    save_folder: pathlib.Path = Field(default=pathlib.Path("./"))
    folder_presets: list[pathlib.Path] = Field(default_factory=list)
    conflict_policy: Annotated[ConflictPolicy, pydantic_enum(ConflictPolicy)] = Field(
        default=ConflictPolicy.PROMPT
    )
    multi_threaded_conversion: bool = Field(default=True)
    open_save_folder_on_completion: bool = Field(default=True)
    auto_set_output_extension: bool = Field(default=True)
    auto_check_for_updates: bool = Field(default=True)


config_path = app_dir.user_config_path / "settings.yml"


if config_path.exists():
    settings = LibreSvipSettings.load(
        app_dir.user_config_path, create_if_missing=True, raise_error_if_failed=False
    )
else:
    settings = LibreSvipSettings()
    settings._settings_dir = app_dir.user_config_path
settings.lyric_replace_rules.setdefault("default", [])


def get_ui_settings() -> LibreSvipBaseUISettings:
    return ui_settings_ctx.get(None) or settings


def save_settings() -> None:
    settings.save()



from pydantic import BaseModel, Field

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableVolumeImportationMixin,
)
from libresvip.utils.translation import gettext_lazy as _


class InputOptions(
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
    EnableVolumeImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    down_sample: int = Field(
        default=32,
        title=_("Average sampling interval for the volume parameter"),
        description=_(
            "The unit is Tick. The larger the value, the smoother the editor; the smaller the value, the more accurate the volume parameter."
        ),
    )
    singer: str = Field(
        default="扇宝",
        title=_("Default singer"),
    )



from libresvip.utils.text import CustomBoundriesBlacklist

unsupported_symbols = CustomBoundriesBlacklist(
    [",", "，", ".", "。", "?", "？", "!", "！"],
    right_boundary="$",
    match_substrings=True,
)


def is_punctuation(c: str) -> bool:
    return unsupported_symbols.is_blacklisted(c)


def get_symbol_removed_lyric(lyric: str) -> str:
    return unsupported_symbols.cleanse_text(lyric)



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import BaseModel, Field

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.option_mixins import EnablePitchImportationMixin
from libresvip.utils.translation import gettext_lazy as _


class BreathOption(Enum):
    IGNORE: Annotated[str, Field(title=_("Ignore all breath notes"))] = "ignore"
    KEEP: Annotated[str, Field(title=_("Keep as normal notes"))] = "keep"


class InputOptions(EnablePitchImportationMixin, BaseModel):
    lyric_encoding: str = Field(
        default="SHIFT_JIS",
        title=_("Lyric text encoding"),
        description=_("Unless the lyrics are garbled, this option should not be changed."),
    )
    breath: BreathOption = Field(
        default=BreathOption.IGNORE,
        title=_("The way to handle breath notes"),
    )


class OutputOptions(BaseModel):
    ticks_per_beat: int = Field(
        default=TICKS_IN_BEAT,
        title=_("Ticks per beat"),
        description=_(
            "Also known as parts per quarter, ticks per quarter note, the number of pulses per quarter note. This setting should not be changed unless you know what it is."
        ),
    )
    lyric_encoding: str = Field(
        default="SHIFT_JIS",
        title=_("Lyric text encoding"),
        description=_("Unless the lyrics are garbled, this option should not be changed."),
    )
